from typing import List, Optional

from pydantic import BaseModel, ConfigDict, Field, model_validator


class HeadMetadata(BaseModel):
    """Metadata for the basic head. It allows you to specify additional
    fields.

    @type postprocessor_path: str | None
    @ivar postprocessor_path: Path to the postprocessor.
    """

    model_config = ConfigDict(extra="allow")

    postprocessor_path: Optional[str] = Field(
        None, description="Path to the postprocessor."
    )


class HeadObjectDetectionMetadata(HeadMetadata):
    """Metadata for the object detection head.

    @type classes: list
    @ivar classes: Names of object classes detected by the model.
    @type n_classes: int
    @ivar n_classes: Number of object classes detected by the model.
    @type iou_threshold: float
    @ivar iou_threshold: Non-max supression threshold limiting boxes
        intersection.
    @type conf_threshold: float
    @ivar conf_threshold: Confidence score threshold above which a
        detected object is considered valid.
    @type max_det: int
    @ivar max_det: Maximum detections per image.
    @type anchors: list
    @ivar anchors: Predefined bounding boxes of different sizes and
        aspect ratios. The innermost lists are length 2 tuples of box
        sizes. The middle lists are anchors for each output. The outmost
        lists go from smallest to largest output.
    """

    classes: List[str] = Field(
        description="Names of object classes recognized by the model."
    )
    n_classes: int = Field(
        description="Number of object classes recognized by the model."
    )
    iou_threshold: float = Field(
        description="Non-max supression threshold limiting boxes intersection."
    )
    conf_threshold: float = Field(
        description="Confidence score threshold above which a detected object is considered valid."
    )
    max_det: int = Field(description="Maximum detections per image.")
    anchors: Optional[List[List[List[float]]]] = Field(
        None,
        description="Predefined bounding boxes of different sizes and aspect ratios. The innermost lists are length 2 tuples of box sizes. The middle lists are anchors for each output. The outmost lists go from smallest to largest output.",
    )


class HeadObjectDetectionSSDMetadata(HeadObjectDetectionMetadata):
    """Metadata for the SSD object detection head.

    @type boxes_outputs: str
    @ivar boxes_outputs: Output name corresponding to predicted bounding
        box coordinates.
    @type scores_outputs: str
    @ivar scores_outputs: Output name corresponding to predicted
        bounding box confidence scores.
    """

    boxes_outputs: str = Field(
        description="Output name corresponding to predicted bounding box coordinates."
    )
    scores_outputs: str = Field(
        description="Output name corresponding to predicted bounding box confidence scores."
    )


class HeadClassificationMetadata(HeadMetadata):
    """Metadata for the classification head.

    @type classes: list
    @ivar classes: Names of object classes classified by the model.
    @type n_classes: int
    @ivar n_classes: Number of object classes classified by the model.
    @type is_softmax: bool
    @ivar is_softmax: True, if output is already softmaxed
    """

    classes: List[str] = Field(
        description="Names of object classes recognized by the model."
    )
    n_classes: int = Field(
        description="Number of object classes recognized by the model."
    )
    is_softmax: bool = Field(
        description="True, if output is already softmaxed."
    )


class HeadSegmentationMetadata(HeadMetadata):
    """Metadata for the segmentation head.

    @type classes: list
    @ivar classes: Names of object classes segmented by the model.
    @type n_classes: int
    @ivar n_classes: Number of object classes segmented by the model.
    @type is_softmax: bool
    @ivar is_softmax: True, if output is already softmaxed
    """

    classes: List[str] = Field(
        description="Names of object classes recognized by the model."
    )
    n_classes: int = Field(
        description="Number of object classes recognized by the model."
    )
    is_softmax: bool = Field(
        description="True, if output is already softmaxed."
    )


class HeadYOLOMetadata(HeadObjectDetectionMetadata, HeadSegmentationMetadata):
    """Metadata for the YOLO head.

    @type yolo_outputs: list
    @ivar yolo_outputs: A list of output names for each of the different
        YOLO grid sizes.
    @type mask_outputs: list | None
    @ivar mask_outputs: A list of output names for each mask output.
    @type protos_outputs: str | None
    @ivar protos_outputs: Output name for the protos.
    @type keypoints_outputs: list | None
    @ivar keypoints_outputs: A list of output names for the keypoints.
    @type angles_outputs: list | None
    @ivar angles_outputs: A list of output names for the angles.
    @type subtype: str
    @ivar subtype: YOLO family decoding subtype (e.g. yolov5, yolov6,
        yolov7 etc.)
    @type n_prototypes: int | None
    @ivar n_prototypes: Number of prototypes per bbox in YOLO instance
        segmnetation.
    @type n_keypoints: int | None
    @ivar n_keypoints: Number of keypoints per bbox in YOLO keypoint
        detection.
    @type is_softmax: bool | None
    @ivar is_softmax: True, if output is already softmaxed in YOLO
        instance segmentation
    """

    yolo_outputs: List[str] = Field(
        description="A list of output names for each of the different YOLO grid sizes."
    )

    # Instance segmentation
    mask_outputs: Optional[List[str]] = Field(
        None, description="A list of output names for each mask output."
    )
    protos_outputs: Optional[str] = Field(
        None, description="Output name for the protos."
    )

    # Keypoint detection
    keypoints_outputs: Optional[List[str]] = Field(
        None, description="A list of output names for the keypoints."
    )

    # OBB detection
    angles_outputs: Optional[List[str]] = Field(
        None, description="A list of output names for the angles."
    )

    subtype: str = Field(
        description="YOLO family decoding subtype (e.g. yolov5, yolov6, yolov7 etc.)."
    )
    n_prototypes: Optional[int] = Field(
        None,
        description="Number of prototypes per bbox in YOLO instance segmnetation.",
    )
    n_keypoints: Optional[int] = Field(
        None,
        description="Number of keypoints per bbox in YOLO keypoint detection.",
    )
    is_softmax: Optional[bool] = Field(
        None,
        description="True, if output is already softmaxed in YOLO instance segmentation.",
    )

    @model_validator(mode="before")
    def validate_task_specific_fields(
        cls,
        values,
    ):
        defined_params = {k for k, v in dict(values).items() if v is not None}

        common_fields = [
            "postprocessor_path",
            "yolo_outputs",
            "mask_outputs",
            "protos_outputs",
            "keypoints_outputs",
            "angles_outputs",
            "subtype",
            "iou_threshold",
            "conf_threshold",
            "max_det",
            "classes",
            "n_classes",
            "anchors",
        ]
        defined_params = defined_params.difference(common_fields)

        required_fields = {
            "object_detection": [],
            "instance_segmentation": [
                "n_prototypes",
                "is_softmax",
            ],
            "keypoint_detection": ["n_keypoints"],
        }

        unsupported_fields = {
            "object_detection": [
                "n_prototypes",
                "n_keypoints",
                "is_softmax",
            ],
            "instance_segmentation": ["n_keypoints"],
            "keypoint_detection": [
                "n_prototypes",
                "is_softmax",
            ],
        }

        tasks = []
        # Extract the task type
        if all(
            [
                field in defined_params
                for field in required_fields.get("instance_segmentation", [])
            ]
        ):
            tasks.append("instance_segmentation")
        if all(
            [
                field in defined_params
                for field in required_fields.get("keypoint_detection", [])
            ]
        ):
            tasks.append("keypoint_detection")
        if all(
            [
                field not in defined_params
                for field in unsupported_fields.get("object_detection", [])
            ]
        ):
            tasks.append("object_detection")

        if len(tasks) == 0:
            raise ValueError(
                "Invalid combination of parameters. No specific task can be inferred."
            )

        for param in defined_params:
            if not any(param in required_fields[task] for task in tasks):
                raise ValueError(
                    f"Invalid combination of parameters. Field {param} is not supported for the tasks {tasks}."
                )

        # Validate Outputs
        defined_params = {k for k, v in dict(values).items() if v is not None}
        common_fields = [
            "postprocessor_path",
            "subtype",
            "iou_threshold",
            "conf_threshold",
            "max_det",
            "classes",
            "n_classes",
            "anchors",
            "n_prototypes",
            "n_keypoints",
            "is_softmax",
        ]
        defined_params = defined_params.difference(common_fields)

        supported_output_params = {
            "instance_segmentation": [
                "yolo_outputs",
                "mask_outputs",
                "protos_outputs",
            ],
            "keypoint_detection": ["yolo_outputs", "keypoints_outputs"],
            "object_detection": ["yolo_outputs"],
        }

        # Check if all required output fields are present
        if not all(
            [
                field in defined_params
                for task in tasks
                for field in supported_output_params[task]
            ]
        ):
            raise ValueError(f"Invalid output fields for tasks {tasks}")

        # Check if all defined fields are supported
        for param in defined_params:
            if param == "angles_outputs" and "object_detection" in tasks:
                continue
            if not any(
                param in supported_output_params[task] for task in tasks
            ):
                raise ValueError(
                    f"Invalid combination of output parameters. Field {param} is not supported for the tasks {tasks}."
                )

        return values



from enum import Enum


class DataType(Enum):
    """Represents all existing data types used in i/o streams of the
    model."""

    INT4 = "int4"
    INT8 = "int8"
    INT32 = "int32"
    INT64 = "int64"
    UINT8 = "uint8"
    UINT16 = "uint16"
    UINT32 = "uint32"
    UINT64 = "uint64"
    FLOAT32 = "float32"
    FLOAT16 = "float16"
    FLOAT64 = "float64"
    BOOLEAN = "boolean"
    STRING = "string"



import pkg_resources

from luxonis_ml.guard_extras import guard_missing_extra

with guard_missing_extra("data"):
    from .augmentations import Augmentations
    from .datasets import (
        DATASETS_REGISTRY,
        BaseDataset,
        DatasetIterator,
        LuxonisComponent,
        LuxonisDataset,
        LuxonisSource,
    )
    from .loaders import (
        LOADERS_REGISTRY,
        BaseLoader,
        Labels,
        LuxonisLoader,
        LuxonisLoaderOutput,
    )
    from .parsers import LuxonisParser
    from .utils.enums import (
        BucketStorage,
        BucketType,
        ImageType,
        LabelType,
        MediaType,
    )


def load_dataset_plugins() -> None:
    """Registers any external dataset BaseDataset class plugins."""
    for entry_point in pkg_resources.iter_entry_points("dataset_plugins"):
        plugin_class = entry_point.load()
        DATASETS_REGISTRY.register_module(module=plugin_class)


def load_loader_plugins() -> None:
    """Registers any external dataset BaseLoader class plugins."""
    for entry_point in pkg_resources.iter_entry_points("loader_plugins"):
        plugin_class = entry_point.load()
        DATASETS_REGISTRY.register_module(module=plugin_class)


load_dataset_plugins()
load_loader_plugins()

__all__ = [
    "Augmentations",
    "BaseDataset",
    "BaseLoader",
    "BucketStorage",
    "BucketType",
    "DatasetIterator",
    "DATASETS_REGISTRY",
    "LOADERS_REGISTRY",
    "ImageType",
    "LabelType",
    "Labels",
    "LuxonisComponent",
    "LuxonisDataset",
    "LuxonisLoader",
    "LuxonisLoaderOutput",
    "LuxonisParser",
    "LuxonisSource",
    "MediaType",
]



from luxonis_ml.data.datasets.utils import (
    add_generator_wrapper,
    rescale_values,
)


def test_rescale_values_keypoints():
    bbox = {"x": 0.1, "y": 0.2, "w": 0.5, "h": 0.5}
    keypoints = [(0.2, 0.4, 2), (0.5, 0.8, 2)]
    expected = [
        (0.2 * 0.5 + 0.1, 0.4 * 0.5 + 0.2, 2),
        (0.5 * 0.5 + 0.1, 0.8 * 0.5 + 0.2, 2),
    ]
    assert rescale_values(bbox, keypoints, "keypoints") == expected


def test_rescale_values_segmentation_polyline():
    bbox = {"x": 0.1, "y": 0.2, "w": 0.5, "h": 0.5}
    segmentation = {"polylines": [[0.2, 0.4], [0.5, 0.8]]}
    expected = [
        (0.2 * 0.5 + 0.1, 0.4 * 0.5 + 0.2),
        (0.5 * 0.5 + 0.1, 0.8 * 0.5 + 0.2),
    ]
    assert rescale_values(bbox, segmentation, "segmentation") == expected


def test_rescale_values_segmentation_rle():
    bbox = {"x": 0.1, "y": 0.2, "w": 0.5, "h": 0.5}
    segmentation = {"rle": {"size": [100, 100], "counts": [40, 0, 60, 0]}}
    # Dummy check: actual implementation requires detailed RLE handling
    result = rescale_values(bbox, segmentation, "segmentation")
    assert isinstance(result, dict)
    assert "height" in result
    assert "width" in result
    assert "counts" in result
    assert result["height"] == 50
    assert result["width"] == 50


def test_rescale_values_invalid_key():
    bbox = {"x": 0.1, "y": 0.2, "w": 0.5, "h": 0.5}
    ann = [0.2, 0.4, 2]
    assert rescale_values(bbox, ann, "invalid_key") is None  # type: ignore


def test_add_generator_wrapper_non_detection():
    def dummy_generator():
        yield {
            "file": "dummy.jpg",
            "annotation": {
                "instance_id": 0,
                "type": "segmentation",
                "class": "person",
                "segmentation": {
                    "polylines": [(0.2, 0.4), (0.5, 0.8), (0.7, 0.9)],
                },
            },
        }

    wrapped_gen = add_generator_wrapper(dummy_generator())
    result = next(wrapped_gen)
    assert isinstance(result, dict)
    assert "file" in result
    assert "annotation" in result
    assert result["annotation"]["type"] == "segmentation"


def test_add_generator_wrapper_missing_scaled_to_boxes():
    def dummy_generator():
        yield {
            "file": "dummy.jpg",
            "annotation": {
                "instance_id": 0,
                "type": "detection",
                "task": "people",
                "class": "person",
                "boundingbox": {"x": 0.1, "y": 0.2, "w": 0.5, "h": 0.5},
                "keypoints": [(0.2, 0.4, 2), (0.5, 0.8, 2)],
            },
        }

    wrapped_gen = add_generator_wrapper(dummy_generator())

    tasks = []
    for record in wrapped_gen:
        assert isinstance(record, dict)
        assert "file" in record
        assert "annotation" in record
        tasks.append(record["annotation"]["task"])

    assert sorted(tasks) == ["people-boundingbox", "people-keypoints"]



from .base_loader import (
    LOADERS_REGISTRY,
    BaseLoader,
    Labels,
    LuxonisLoaderOutput,
)
from .luxonis_loader import LuxonisLoader

__all__ = [
    "BaseLoader",
    "Labels",
    "LuxonisLoader",
    "LuxonisLoaderOutput",
    "LOADERS_REGISTRY",
]



from importlib.metadata import version

import typer

from luxonis_ml.utils import setup_logging

app = typer.Typer(
    name="Luxonis ML CLI",
    add_completion=True,
    pretty_exceptions_show_locals=False,
)
setup_logging(use_rich=True)

try:
    from luxonis_ml.data.__main__ import app as data_app

    app.add_typer(data_app, name="data", help="Dataset utilities.")
except ImportError:
    pass

try:
    from luxonis_ml.utils.__main__ import app as utils_app

    app.add_typer(utils_app, name="fs", help="Filesystem utilities.")
except ImportError:
    pass

try:
    from luxonis_ml.nn_archive.__main__ import app as utils_app

    app.add_typer(utils_app, name="archive", help="NN Archive utilities.")
except ImportError:
    pass


def version_callback(value: bool):
    if value:
        print(f"LuxonisML: {version(__package__)}")
        raise typer.Exit()


@app.callback()
def main(
    _: bool = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Show version and exit.",
    ),
):
    # Do other global stuff, handle other global options here
    return


if __name__ == "__main__":
    app()



import glob
import json
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import cv2
import numpy as np

from luxonis_ml.data import DatasetIterator

from .base_parser import BaseParser, ParserOutput


class SOLOParser(BaseParser):
    """Parses directory with SOLO annotations to LDF.

    Expected format::

        dataset_dir/
        ├── train/
        │   ├── metadata.json
        │   ├── sensor_definitions.json
        │   ├── annotation_definitions.json
        │   ├── metric_definitions.json
        │   └── sequence.<SequenceNUM>/
        │       ├── step<StepNUM>.camera.jpg
        │       ├── step<StepNUM>.frame_data.json
        │       └── (OPTIONAL: step<StepNUM>.camera.semantic segmentation.jpg)
        ├── valid/
        └── test/

    This is the default format returned by Unity simulation engine.
    """

    @staticmethod
    def validate_split(split_path: Path) -> Optional[Dict[str, Any]]:
        """Validates if a split subdirectory is in an expected format.

        @type split_path: Path
        @param split_path: Path to split directory.
        @rtype: Optional[Dict[str, Any]]
        @return: Dictionary with kwargs to pass to L{from_split} method
            or C{None} if the split is not in the expected format.
        """
        if not split_path.exists():
            return None
        # check if all json files are present
        for json_fname in [
            "annotation_definitions.json",
            "metadata.json",
            "metric_definitions.json",
            "sensor_definitions.json",
        ]:
            json_path = next(split_path.glob(json_fname), None)
            if not json_path:
                return None
        with open(os.path.join(split_path, "metadata.json")) as json_file:
            metadata_dict = json.load(json_file)
        # check if all sequences are present
        total_sequences_expected = metadata_dict["totalSequences"]
        total_sequences = len(
            [
                d
                for d in glob.glob(os.path.join(split_path, "sequence*"))
                if os.path.isdir(d)
            ]
        )
        if not total_sequences == total_sequences_expected:
            return None
        return {"split_path": split_path}

    @staticmethod
    def validate(dataset_dir: Path) -> bool:
        """Validates if the dataset is in an expected format.

        @type dataset_dir: Path
        @param dataset_dir: Path to source dataset directory.
        @rtype: bool
        @return: True if the dataset is in the expected format.
        """
        for split in ["train", "valid", "test"]:
            split_path = dataset_dir / split
            if SOLOParser.validate_split(split_path) is None:
                return False
        return True

    def from_dir(
        self,
        dataset_dir: Path,
    ) -> Tuple[List[str], List[str], List[str]]:
        """Parses all present data to L{LuxonisDataset} format.

        @type dataset_dir: str
        @param dataset_dir: Path to source dataset directory.
        @rtype: Tuple[List[str], List[str], List[str]]
        @return: Tuple with added images for train, valid and test
            splits.
        """

        added_train_imgs = self._parse_split(split_path=dataset_dir / "train")
        added_valid_imgs = self._parse_split(split_path=dataset_dir / "valid")
        added_test_imgs = self._parse_split(split_path=dataset_dir / "test")

        return added_train_imgs, added_valid_imgs, added_test_imgs

    def from_split(
        self,
        split_path: Path,
    ) -> ParserOutput:
        """Parses data in a split subdirectory from SOLO format to
        L{LuxonisDataset} format.

        @type split_path: Path
        @param split_path: Path to directory with sequences of images
            and annotations.
        @rtype: L{ParserOutput}
        @return: C{LuxonisDataset} generator, list of class names,
            skeleton dictionary for keypoints and list of added images.
        """

        if not os.path.exists(split_path):
            raise Exception(f"{split_path} path non-existent.")

        annotation_definitions_path = os.path.join(
            split_path, "annotation_definitions.json"
        )
        if os.path.exists(annotation_definitions_path):
            with open(annotation_definitions_path) as json_file:
                annotation_definitions_dict = json.load(json_file)
        else:
            raise Exception(
                f"{annotation_definitions_path} path non-existent."
            )

        annotation_types = self._get_solo_annotation_types(
            annotation_definitions_dict
        )

        class_names = self._get_solo_bbox_class_names(
            annotation_definitions_dict
        )
        # TODO: We make an assumption here that bbox class_names are also valid for all other annotation types in the dataset. Is this OK?
        # TODO: Can we imagine a case where classes between annotation types are different? Which class names to return in this case?
        if class_names == []:
            raise Exception("No class_names identified. ")

        keypoint_labels = self._get_solo_keypoint_names(
            annotation_definitions_dict
        )

        skeletons = {
            class_name: {"labels": keypoint_labels}
            for class_name in class_names
        }
        # TODO: setting skeletons by assigning all keypoint names to each class_name. Is this OK?
        # if NOT, set them manually with LuxonisDataset.set_skeletons() as SOLO format does not
        # encode which keypoint_name belongs to which class

        def generator() -> DatasetIterator:
            for dir_path, dir_names, _ in os.walk(split_path):
                for dir_name in dir_names:
                    if not dir_name.startswith("sequence"):
                        continue

                    sequence_path = os.path.join(dir_path, dir_name)

                    for frame_path in glob.glob(
                        os.path.join(sequence_path, "*.frame_data.json")
                    ):  # single sequence can have multiple steps
                        if not os.path.exists(frame_path):
                            raise FileNotFoundError(
                                f"{frame_path} not existent."
                            )
                        with open(frame_path) as f:
                            frame = json.load(f)

                        for capture in frame["captures"]:
                            img_fname = capture["filename"]
                            img_w, img_h = capture["dimension"]
                            annotations = capture["annotations"]
                            img_path = os.path.join(sequence_path, img_fname)
                            if not os.path.exists(img_path):
                                raise FileNotFoundError(
                                    f"{img_path} not existent."
                                )

                            if "BoundingBox2DAnnotation" in annotation_types:
                                for anno in annotations:
                                    if anno["@type"].endswith(
                                        "BoundingBox2DAnnotation"
                                    ):
                                        bbox_annotations = anno["values"]

                                for bbox_annotation in bbox_annotations:
                                    class_name = bbox_annotation["labelName"]

                                    origin = bbox_annotation["origin"]
                                    dimension = bbox_annotation["dimension"]
                                    xmin, ymin = origin
                                    bbox_w, bbox_h = dimension

                                    yield {
                                        "file": img_path,
                                        "annotation": {
                                            "type": "boundingbox",
                                            "class": class_name,
                                            "x": xmin / img_w,
                                            "y": ymin / img_h,
                                            "w": bbox_w / img_w,
                                            "h": bbox_h / img_h,
                                        },
                                    }

                            if (
                                "SemanticSegmentationAnnotation"
                                in annotation_types
                            ):
                                for anno in annotations:
                                    if anno["@type"].endswith(
                                        "SemanticSegmentationAnnotation"
                                    ):
                                        sseg_annotations = anno

                                mask_fname = sseg_annotations["filename"]
                                mask_path = os.path.join(
                                    sequence_path, mask_fname
                                )
                                mask = cv2.imread(mask_path)

                                for instance in sseg_annotations["instances"]:
                                    class_name = instance["labelName"]
                                    r, g, b, _ = instance["pixelValue"]
                                    curr_mask = np.zeros_like(mask)
                                    curr_mask[
                                        np.all(mask == [b, g, r], axis=2)
                                    ] = 1
                                    curr_mask = np.max(
                                        curr_mask, axis=2
                                    )  # 3D->2D

                                    yield {
                                        "file": img_path,
                                        "annotation": {
                                            "type": "mask",
                                            "class": class_name,
                                            "mask": curr_mask,
                                        },
                                    }

                            if "KeypointAnnotation" in annotation_types:
                                for anno in annotations:
                                    if anno["@type"].endswith(
                                        "KeypointAnnotation"
                                    ):
                                        keypoint_annotations = anno["values"]

                                for (
                                    keypoints_annotation
                                ) in keypoint_annotations:
                                    label_id = keypoints_annotation["labelId"]
                                    keypoints = []
                                    for keypoint in keypoints_annotation[
                                        "keypoints"
                                    ]:
                                        x, y = keypoint["location"]
                                        visibility = keypoint["state"]
                                        keypoints.append(
                                            (x / img_w, y / img_h, visibility)
                                        )

                                    class_name = class_names[label_id]

                                    yield {
                                        "file": img_path,
                                        "annotation": {
                                            "type": "keypoints",
                                            "class": class_name,
                                            "keypoints": keypoints,
                                        },
                                    }

        added_images = self._get_added_images(generator())

        return (
            generator(),
            class_names,
            skeletons,
            added_images,
        )

    def _get_solo_annotation_types(
        self, annotation_definitions_dict: dict
    ) -> list:
        """List all annotation types present in the dataset.

        @type annotation_definitions_dict: dict
        @param annotation_definitions_dict: annotation_definitions.json read as dict.
        @rtype: List[str]
        @return: List of annotation types (e.g. ["BoundingBox2DAnnotation", "SemanticSegmentationAnnotation", ...]).
        """

        annotation_types = []
        for definition in annotation_definitions_dict["annotationDefinitions"]:
            annotation_types.append(
                definition["@type"].replace("type.unity.com/unity.solo.", "")
            )
        return annotation_types

    def _get_solo_bbox_class_names(self, annotation_definitions_dict):
        """List class names for BoundingBox2DAnnotation type.

        @type annotation_definitions_dict: dict
        @param annotation_definitions_dict: annotation_definitions.json read as dict.
        @rtype: List[str]
        @return: List of class names (e.g. ["car", "motorbike", ...]).
        """

        class_names = []
        for definition in annotation_definitions_dict["annotationDefinitions"]:
            annotation_type = definition["@type"].replace(
                "type.unity.com/unity.solo.", ""
            )
            if annotation_type == "BoundingBox2DAnnotation":
                names = [spec["label_name"] for spec in definition["spec"]]
                ids = [spec["label_id"] for spec in definition["spec"]]
                class_names = [c for _, c in sorted(zip(ids, names))]
        return class_names

    def _get_solo_keypoint_names(self, annotation_definitions_dict):
        """List keypoint labels for all classes.

        @type annotation_definitions_dict: dict
        @param annotation_definitions_dict: annotation_definitions.json read as dict.
        @rtype: List[str]
        @return: List of keypoint labels (e.g. ["wheel_front_motorbike", "wheel_back_motorbike", ...]).
        """

        keypoint_labels = []
        for definition in annotation_definitions_dict["annotationDefinitions"]:
            annotation_type = definition["@type"].replace(
                "type.unity.com/unity.solo.", ""
            )
            if annotation_type == "KeypointAnnotation":
                keypoints = definition["template"]["keypoints"]
                labels = [keypoint["label"] for keypoint in keypoints]
                ids = [keypoint["index"] for keypoint in keypoints]
                keypoint_labels = [c for _, c in sorted(zip(ids, labels))]
        return keypoint_labels



import json
import logging
from abc import ABC, abstractmethod
from datetime import datetime, timezone
from typing import (
    Any,
    ClassVar,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    TypedDict,
    Union,
)

import numpy as np
import pycocotools.mask as mask_util
from PIL import Image, ImageDraw
from pydantic import (
    ConfigDict,
    Field,
    field_serializer,
    field_validator,
    model_validator,
)
from pydantic.types import FilePath, NonNegativeInt, PositiveInt
from typing_extensions import Annotated, TypeAlias

from luxonis_ml.utils import BaseModelExtraForbid

from ..utils.enums import LabelType

logger = logging.getLogger(__name__)

KeypointVisibility: TypeAlias = Literal[0, 1, 2]
NormalizedFloat: TypeAlias = Annotated[float, Field(ge=0, le=1)]
"""C{NormalizedFloat} is a float that is restricted to the range [0,
1]."""

ParquetDict = TypedDict(
    "ParquetDict",
    {
        "file": str,
        "type": str,
        "created_at": datetime,
        "class": str,
        "instance_id": int,
        "task": str,
        "annotation": str,
    },
)


def load_annotation(name: str, data: Dict[str, Any]) -> "Annotation":
    return {
        "ClassificationAnnotation": ClassificationAnnotation,
        "BBoxAnnotation": BBoxAnnotation,
        "KeypointAnnotation": KeypointAnnotation,
        "RLESegmentationAnnotation": RLESegmentationAnnotation,
        "PolylineSegmentationAnnotation": PolylineSegmentationAnnotation,
        "MaskSegmentationAnnotation": MaskSegmentationAnnotation,
        "ArrayAnnotation": ArrayAnnotation,
        "LabelAnnotation": LabelAnnotation,
    }[name](**data)


class Annotation(ABC, BaseModelExtraForbid):
    """Base class for an annotation.

    @type task: str
    @ivar task: The task name. By default it is the string
        representation of the L{LabelType}.
    @type class_: str
    @ivar class_: The class name for the annotation.
    @type instance_id: int
    @ivar instance_id: The instance id of the annotation. This
        determines the order in which individual instances are loaded in
        L{luxonis_ml.data.LuxonisLoader}.
    @type _label_type: ClassVar[LabelType]
    @ivar _label_type: The label type of the annotation.
    """

    _label_type: ClassVar[LabelType]

    task: str
    class_: str = Field("", alias="class")
    instance_id: int = -1

    @model_validator(mode="before")
    @classmethod
    def validate_task(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if "task" not in values:
            values = {**values, "task": cls._label_type.value}
        return values

    def get_value(self) -> Dict[str, Any]:
        """Converts the annotation to a dictionary that can be saved to
        a parquet file."""
        return self.dict(
            exclude={"class_", "class_id", "instance_id", "task", "type_"}
        )

    @staticmethod
    @abstractmethod
    def combine_to_numpy(
        annotations: List["Annotation"],
        class_mapping: Dict[str, int],
        height: int,
        width: int,
    ) -> np.ndarray:
        """Combines multiple instance annotations into a single numpy
        array.

        @type annotations: List[Annotation]
        @param annotations: List of annotations to combine.
        @type class_mapping: Dict[str, int]
        @param class_mapping: Mapping of class names to class indices.
        @type height: int
        @param height: The height of the image.
        @type width: int
        @param width: The width of the image.
        """
        pass


class ClassificationAnnotation(Annotation):
    _label_type = LabelType.CLASSIFICATION
    type_: Literal["classification"] = Field("classification", alias="type")

    @staticmethod
    def combine_to_numpy(
        annotations: List["ClassificationAnnotation"],
        class_mapping: Dict[str, int],
        **_,
    ) -> np.ndarray:
        classify_vector = np.zeros(len(class_mapping))
        for ann in annotations:
            class_ = class_mapping.get(ann.class_, 0)
            classify_vector[class_] = 1
        return classify_vector


class BBoxAnnotation(Annotation):
    """Bounding box annotation.

    Values are normalized based on the image size.

    @type x: float
    @ivar x: The top-left x coordinate of the bounding box. Normalized
        to M{[0, 1]}.
    @type y: float
    @ivar y: The top-left y coordinate of the bounding box. Normalized
        to M{[0, 1]}.
    @type w: float
    @ivar w: The width of the bounding box. Normalized to M{[0, 1]}.
    @type h: float
    @ivar h: The height of the bounding box. Normalized to M{[0, 1]}.
    """

    type_: Literal["boundingbox"] = Field("boundingbox", alias="type")

    x: NormalizedFloat
    y: NormalizedFloat
    w: NormalizedFloat
    h: NormalizedFloat

    _label_type = LabelType.BOUNDINGBOX

    @model_validator(mode="before")
    @classmethod
    def validate_values(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        warn = False
        for key in ["x", "y", "w", "h"]:
            if values[key] < -2 or values[key] > 2:
                raise ValueError(
                    "BBox annotation has value outside of automatic clipping range ([-2, 2]). "
                    "Values should be normalized based on image size to range [0, 1]."
                )
            if not (0 <= values[key] <= 1):
                warn = True
                values[key] = max(0, min(1, values[key]))
        if warn:
            logger.warning(
                "BBox annotation has values outside of [0, 1] range. Clipping them to [0, 1]."
            )

        # cliping done in function instead of separate model validator so
        # order of execution is explicitly defined
        values = cls.clip_sum(values)
        return values

    @classmethod
    def clip_sum(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if values["x"] + values["w"] > 1:
            values["w"] = 1 - values["x"]
            logger.warning(
                "BBox annotation has x + width > 1. Clipping width so the sum is 1."
            )
        if values["y"] + values["h"] > 1:
            values["h"] = 1 - values["y"]
            logger.warning(
                "BBox annotation has y + height > 1. Clipping height so the sum is 1."
            )
        return values

    def to_numpy(self, class_mapping: Dict[str, int]) -> np.ndarray:
        class_ = class_mapping.get(self.class_, 0)
        return np.array([class_, self.x, self.y, self.w, self.h])

    @staticmethod
    def combine_to_numpy(
        annotations: List["BBoxAnnotation"], class_mapping: Dict[str, int], **_
    ) -> np.ndarray:
        boxes = np.zeros((len(annotations), 5))
        for i, ann in enumerate(annotations):
            boxes[i] = ann.to_numpy(class_mapping)
        return boxes


class KeypointAnnotation(Annotation):
    """Keypoint annotation.

    Values are normalized to M{[0, 1]} based on the image size.

    @type keypoints: List[Tuple[float, float, L{KeypointVisibility}]]
    @ivar keypoints: List of keypoints. Each keypoint is a tuple of (x, y, visibility).
        x and y are normalized to M{[0, 1]}. visibility is one of M{0}, M{1}, or M{2} where:
            - 0: Not visible / not labeled
            - 1: Occluded
            - 2: Visible
    """

    type_: Literal["keypoints"] = Field("keypoints", alias="type")

    keypoints: List[
        Tuple[NormalizedFloat, NormalizedFloat, KeypointVisibility]
    ]

    _label_type = LabelType.KEYPOINTS

    @model_validator(mode="before")
    @classmethod
    def validate_values(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        warn = False
        for i, keypoint in enumerate(values["keypoints"]):
            if (keypoint[0] < -2 or keypoint[0] > 2) or (
                keypoint[1] < -2 or keypoint[1] > 2
            ):
                raise ValueError(
                    "Keypoint annotation has value outside of automatic clipping range ([-2, 2]). "
                    "Values should be normalized based on image size to range [0, 1]."
                )
            new_keypoint = list(keypoint)
            if not (0 <= keypoint[0] <= 1):
                new_keypoint[0] = max(0, min(1, keypoint[0]))
                warn = True
            if not (0 <= keypoint[1] <= 1):
                new_keypoint[1] = max(0, min(1, keypoint[1]))
                warn = True
            values["keypoints"][i] = tuple(new_keypoint)

        if warn:
            logger.warning(
                "Keypoint annotation has values outside of [0, 1] range. Clipping them to [0, 1]."
            )
        return values

    def to_numpy(self, class_mapping: Dict[str, int]) -> np.ndarray:
        class_ = class_mapping.get(self.class_, 0)
        kps = np.array(self.keypoints).reshape((-1, 3)).astype(np.float32)
        return np.concatenate([[class_], kps.flatten()])

    @staticmethod
    def combine_to_numpy(
        annotations: List["KeypointAnnotation"],
        class_mapping: Dict[str, int],
        **_,
    ) -> np.ndarray:
        keypoints = np.zeros(
            (len(annotations), len(annotations[0].keypoints) * 3 + 1)
        )
        for i, ann in enumerate(annotations):
            keypoints[i] = ann.to_numpy(class_mapping)
        return keypoints


class SegmentationAnnotation(Annotation):
    """Base class for segmentation annotations."""

    _label_type = LabelType.SEGMENTATION

    @abstractmethod
    def to_numpy(
        self, class_mapping: Dict[str, int], width: int, height: int
    ) -> np.ndarray:
        """Converts the annotation to a numpy array."""
        pass

    @staticmethod
    def combine_to_numpy(
        annotations: List["SegmentationAnnotation"],
        class_mapping: Dict[str, int],
        height: int,
        width: int,
    ) -> np.ndarray:
        seg = np.zeros((len(class_mapping), height, width), dtype=np.uint8)

        masks = np.stack(
            [ann.to_numpy(class_mapping, width, height) for ann in annotations]
        )
        classes = np.array(
            [class_mapping.get(ann.class_, 0) for ann in annotations]
        )

        for i, class_ in enumerate(classes):
            seg[class_, ...] = np.maximum(
                seg[class_, ...], masks[i].astype(np.uint8)
            )

        return seg


class RLESegmentationAnnotation(SegmentationAnnotation):
    """Run-length encoded segmentation mask.

    @type height: int
    @ivar height: The height of the segmentation mask.

    @type width: int
    @ivar width: The width of the segmentation mask.

    @type counts: Union[List[int], bytes]
    @ivar counts: The run-length encoded mask.
        This can be a list of integers or a byte string.

    @see: U{Run-length encoding<https://en.wikipedia.org/wiki/Run-length_encoding>}
    """

    type_: Literal["rle"] = Field("rle", alias="type")

    height: PositiveInt
    width: PositiveInt
    counts: Union[List[NonNegativeInt], bytes]

    def get_value(self) -> Dict[str, Any]:
        if isinstance(self.counts, bytes):
            return {
                "height": self.height,
                "width": self.width,
                "counts": self.counts.decode("utf-8"),
            }

        rle: Any = mask_util.frPyObjects(
            {
                "counts": self.counts,
                "size": [self.height, self.width],
            },
            self.height,
            self.width,
        )

        return {
            "height": rle["size"][0],
            "width": rle["size"][1],
            "counts": rle["counts"].decode("utf-8"),
        }

    def to_numpy(
        self, _: Dict[str, int], width: int, height: int
    ) -> np.ndarray:
        assert isinstance(self.counts, bytes)
        return mask_util.decode(
            {"counts": self.counts, "size": [height, width]}
        ).astype(np.bool_)


class MaskSegmentationAnnotation(SegmentationAnnotation):
    """Pixel-wise binary segmentation mask.

    @type mask: npt.NDArray[np.bool_]
    @ivar mask: The segmentation mask as a numpy array. The mask must be
        2D and must be castable to a boolean array.
    """

    type_: Literal["mask"] = Field("mask", alias="type")
    mask: np.ndarray

    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True)

    @model_validator(mode="before")
    @classmethod
    def _convert_rle(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if "mask" in values:
            return values

        if (
            "width" not in values
            or "height" not in values
            or "counts" not in values
        ):
            raise ValueError(
                "MaskSegmentationAnnotation must have either "
                "'mask' or 'width', 'height', and 'counts'"
            )

        width: int = values.pop("width")
        height: int = values.pop("height")
        counts: str = values.pop("counts")

        values["mask"] = mask_util.decode(
            {
                "counts": counts.encode("utf-8"),
                "size": [height, width],
            }
        ).astype(np.bool_)
        return values

    @field_validator("mask", mode="after")
    @staticmethod
    def _validate_shape(mask: np.ndarray) -> np.ndarray:
        if mask.ndim != 2:
            raise ValueError("Mask must be a 2D array")
        return mask

    @field_validator("mask", mode="before")
    @staticmethod
    def _validate_mask(mask: Any) -> np.ndarray:
        if not isinstance(mask, np.ndarray):
            raise ValueError("Mask must be a numpy array")
        return mask.astype(np.bool_)

    def get_value(self) -> Dict[str, Any]:
        mask = np.asfortranarray(self.mask.astype(np.uint8))
        rle = mask_util.encode(mask)

        return {
            "height": rle["size"][0],
            "width": rle["size"][1],
            "counts": rle["counts"].decode("utf-8"),  # type: ignore
        }

    def to_numpy(self, *_) -> np.ndarray:
        return self.mask


class PolylineSegmentationAnnotation(SegmentationAnnotation):
    """Polyline segmentation mask.

    @type points: List[Tuple[float, float]]
    @ivar points: List of points that define the polyline. Each point is
        a tuple of (x, y). x and y are normalized to M{[0, 1]} based on
        the image size.
    """

    type_: Literal["polyline"] = Field("polyline", alias="type")

    points: List[Tuple[NormalizedFloat, NormalizedFloat]] = Field(min_length=3)

    @model_validator(mode="before")
    @classmethod
    def validate_values(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        warn = False
        for i, point in enumerate(values["points"]):
            if (point[0] < -2 or point[0] > 2) or (
                point[1] < -2 or point[1] > 2
            ):
                raise ValueError(
                    "Polyline annotation has value outside of automatic clipping range ([-2, 2]). "
                    "Values should be normalized based on image size to range [0, 1]."
                )
            new_point = list(point)
            if not (0 <= point[0] <= 1):
                new_point[0] = max(0, min(1, point[0]))
                warn = True
            if not (0 <= point[1] <= 1):
                new_point[1] = max(0, min(1, point[1]))
                warn = True
            values["points"][i] = tuple(new_point)

        if warn:
            logger.warning(
                "Polyline annotation has values outside of [0, 1] range. Clipping them to [0, 1]."
            )
        return values

    def to_numpy(
        self, _: Dict[str, int], width: int, height: int
    ) -> np.ndarray:
        polyline = [
            (round(x * width), round(y * height)) for x, y in self.points
        ]
        mask = Image.new("L", (width, height), 0)
        draw = ImageDraw.Draw(mask)
        draw.polygon(polyline, fill=1, outline=1)
        return np.array(mask).astype(np.bool_)


class ArrayAnnotation(Annotation):
    """A custom unspecified annotation that is an arbitrary numpy array.

    All instances of this annotation must have the same shape.

    @type path: FilePath
    @ivar path: The path to the numpy array saved as a C{.npy} file.
    """

    type_: Literal["array"] = Field("array", alias="type")

    path: FilePath

    _label_type = LabelType.ARRAY

    @staticmethod
    def combine_to_numpy(
        annotations: List["ArrayAnnotation"],
        class_mapping: Dict[str, int],
        **_,
    ) -> np.ndarray:
        out_arr = np.zeros(
            (
                len(annotations),
                len(class_mapping),
                *np.load(annotations[0].path).shape,
            )
        )
        for i, ann in enumerate(annotations):
            class_ = class_mapping.get(ann.class_, 0)
            out_arr[i, class_] = np.load(ann.path)
        return out_arr

    @field_serializer("path")
    def serialize_path(self, value: FilePath) -> str:
        return str(value)


class LabelAnnotation(Annotation):
    """A custom unspecified annotation with a single primitive value.

    @type value: Union[bool, int, float, str]
    @ivar value: The value of the annotation.
    """

    type_: Literal["label"] = Field("label", alias="type")

    value: Union[bool, int, float, str]

    _label_type = LabelType.LABEL

    @staticmethod
    def combine_to_numpy(
        annotations: List["LabelAnnotation"],
        class_mapping: Dict[str, int],
        **_,
    ) -> np.ndarray:
        out_arr = np.zeros((len(annotations), len(class_mapping))).astype(
            type(annotations[0].value)
        )
        for i, ann in enumerate(annotations):
            class_ = class_mapping.get(ann.class_, 0)
            out_arr[i, class_] = ann.value
        return out_arr


class DatasetRecord(BaseModelExtraForbid):
    """A record of an image and its annotation.

    @type file: FilePath
    @ivar file: A path to the image.
    @type annotation: Optional[Annotation]
    @ivar annotation: The annotation for the image.
    """

    file: FilePath
    annotation: Optional[
        Union[
            ClassificationAnnotation,
            BBoxAnnotation,
            KeypointAnnotation,
            RLESegmentationAnnotation,
            PolylineSegmentationAnnotation,
            MaskSegmentationAnnotation,
            ArrayAnnotation,
            LabelAnnotation,
        ]
    ] = Field(None, discriminator="type_")

    def to_parquet_dict(self) -> ParquetDict:
        """Converts an annotation to a dictionary for writing to a
        parquet file.

        @rtype: L{ParquetDict}
        @return: A dictionary of annotation data.
        """

        value = (
            self.annotation.get_value() if self.annotation is not None else {}
        )
        json_value = json.dumps(value)
        return {
            "file": self.file.name,
            "type": self.annotation.__class__.__name__,
            "created_at": datetime.now(timezone.utc),
            "class": (
                self.annotation.class_ or ""
                if self.annotation is not None
                else ""
            ),
            "instance_id": (
                self.annotation.instance_id or -1
                if self.annotation is not None
                else -1
            ),
            "task": self.annotation.task
            if self.annotation is not None
            else "",
            "annotation": json_value,
        }



from abc import ABC, abstractmethod
from typing import Dict, Iterator, List, Optional, Sequence, Tuple, Type, Union

from typing_extensions import TypeAlias

from luxonis_ml.data.datasets.annotation import DatasetRecord
from luxonis_ml.data.datasets.source import LuxonisSource
from luxonis_ml.utils import AutoRegisterMeta, Registry
from luxonis_ml.utils.filesystem import PathType

DATASETS_REGISTRY: Registry[Type["BaseDataset"]] = Registry(name="datasets")


DatasetIterator: TypeAlias = Iterator[Union[dict, DatasetRecord]]


class BaseDataset(
    ABC, metaclass=AutoRegisterMeta, registry=DATASETS_REGISTRY, register=False
):
    """Base abstract dataset class for managing datasets in the Luxonis
    MLOps ecosystem."""

    @property
    @abstractmethod
    def identifier(self) -> str:
        """The unique identifier for the dataset.

        @type: str
        """
        pass

    @abstractmethod
    def get_tasks(self) -> List[str]:
        """Returns the list of tasks in the dataset.

        @rtype: List[str]
        @return: List of task names.
        """
        pass

    @abstractmethod
    def update_source(self, source: LuxonisSource) -> None:
        """Updates underlying source of the dataset with a new
        LuxonisSource.

        @type source: L{LuxonisSource}
        @param source: The new C{LuxonisSource} to replace the old one.
        """
        pass

    @abstractmethod
    def set_classes(
        self, classes: List[str], task: Optional[str] = None
    ) -> None:
        """Sets the names of classes for the dataset. This can be across
        all CV tasks or certain tasks.

        @type classes: List[str]
        @param classes: List of class names to set.
        @type task: Optional[str]
        @param task: Optionally specify the task where these classes
            apply.
        """
        pass

    @abstractmethod
    def get_classes(self) -> Tuple[List[str], Dict[str, List[str]]]:
        """Gets overall classes in the dataset and classes according to
        computer vision task.

        @rtype: Tuple[List[str], Dict]
        @return: A combined list of classes for all tasks and a
            dictionary mapping tasks to the classes used in each task.
        """
        pass

    @abstractmethod
    def set_skeletons(
        self,
        labels: Optional[List[str]] = None,
        edges: Optional[List[Tuple[int, int]]] = None,
        task: Optional[str] = None,
    ) -> None:
        """Sets the semantic structure of keypoint skeletons for the
        classes that use keypoints.

        Example::

            dataset.set_skeletons(
                labels=["right hand", "right shoulder", ...],
                edges=[[0, 1], [4, 5], ...]
            )

        @type labels: Optional[List[str]]
        @param labels: List of keypoint names.
        @type edges: Optional[List[Tuple[int, int]]]
        @param edges: List of edges between keypoints.
        @type task: Optional[str]
        @param task: Optionally specify the task where these skeletons apply.
            If not specified, the skeletons are set for all tasks that use keypoints.
        """
        pass

    @abstractmethod
    def get_skeletons(
        self,
    ) -> Dict[str, Tuple[List[str], List[Tuple[int, int]]]]:
        """Returns the dictionary defining the semantic skeleton for
        each class using keypoints.

        @rtype: Dict[str, Tuple[List[str], List[Tuple[int, int]]]]
        @return: For each task, a tuple containing a list of keypoint
            names and a list of edges between the keypoints.
        """
        pass

    @abstractmethod
    def add(
        self, generator: DatasetIterator, batch_size: int = 1_000_000
    ) -> None:
        """Write annotations to parquet files.

        @type generator: L{DatasetIterator}
        @param generator: A Python iterator that yields either instances
            of C{DatasetRecord} or a dictionary that can be converted to
            C{DatasetRecord}.
        @type batch_size: int
        @param batch_size: The number of annotations generated before
            processing. This can be set to a lower value to reduce
            memory usage.
        """
        pass

    @abstractmethod
    def make_splits(
        self,
        splits: Optional[
            Union[
                Dict[str, Sequence[PathType]],
                Dict[str, float],
                Tuple[float, float, float],
            ]
        ] = None,
        *,
        ratios: Optional[
            Union[Dict[str, float], Tuple[float, float, float]]
        ] = None,
        definitions: Optional[Dict[str, List[PathType]]] = None,
        replace_old_splits: bool = False,
    ) -> None:
        """Generates splits for the dataset.

        @type splits: Optional[Union[Dict[str, Sequence[PathType]], Dict[str, float], Tuple[float, float, float]]]
        @param splits: A dictionary of splits or a tuple of ratios for train, val, and test splits. Can be one of:
            - A dictionary of splits with keys as split names and values as lists of filepaths
            - A dictionary of splits with keys as split names and values as ratios
            - A 3-tuple of ratios for train, val, and test splits
        @type ratios: Optional[Union[Dict[str, float], Tuple[float, float, float]]]
        @param ratios: Deprecated! A dictionary of splits with keys as split names and values as ratios.
        @type definitions: Optional[Dict[str, List[PathType]]]
        @param definitions: Deprecated! A dictionary of splits with keys as split names and values as lists of filepaths.
        @type replace_old_splits: bool
        @param replace_old_splits: Whether to remove old splits and generate new ones. If set to False, only new files will be added to the splits. Default is False.
        """
        pass

    @abstractmethod
    def delete_dataset(self) -> None:
        """Deletes all local files belonging to the dataset."""
        pass

    @staticmethod
    @abstractmethod
    def exists(dataset_name: str) -> bool:
        """Checks whether a dataset exists.

        @warning: For offline mode only.
        @type dataset_name: str
        @param dataset_name: Name of the dataset
        @rtype: bool
        @return: Whether the dataset exists
        """
        pass



import tarfile
from pathlib import Path
from typing import List

from luxonis_ml.utils.filesystem import PathType


def is_nn_archive(path: PathType) -> bool:
    """Check if the given path is a valid NN archive file.

    @type path: PathType
    @param path: Path to the file to check.
    @rtype: bool
    @return: True if the file is a valid NN archive file, False
        otherwise.
    """
    path = Path(path)

    if not path.is_file():
        return False

    if not tarfile.is_tarfile(path):
        return False

    with tarfile.open(path, "r") as tar:
        if "config.json" not in tar.getnames():
            return False

    return True


def infer_layout(shape: List[int]) -> str:
    """Infers a layout for the given shape.

    Tries to guess most common layouts for the given shape pattern.
    Otherwise, uses the first free letter of the alphabet for each dimension.

    Example::
        >>> make_default_layout([1, 3, 256, 256])
        >>> "NCHW"
        >>> make_default_layout([1, 19, 7, 8])
        >>> "NABC"
    """
    layout = []
    i = 0
    if shape[0] == 1:
        layout.append("N")
        i += 1
    if len(shape) - i == 3:
        if shape[i] < shape[i + 1] and shape[i] < shape[i + 2]:
            return "".join(layout + ["C", "H", "W"])
        elif shape[-1] < shape[-2] and shape[-1] < shape[-3]:
            return "".join(layout + ["H", "W", "C"])
    i = 0
    while len(layout) < len(shape):
        # Starting with "C" for more sensible defaults
        letter = chr(ord("A") + i + 2)
        if ord(letter) > ord("Z"):
            raise ValueError(
                f"Too many dimensions ({len(shape)}) for automatic layout."
            )

        if letter not in layout:
            layout.append(letter)
        i += 1
    return "".join(layout)



from typing import Any, Dict, List, Optional, Tuple

import weaviate
import weaviate.classes as wvc
from weaviate.collections.classes.filters import FilterMetadata

from luxonis_ml.embeddings.utils.vectordb import VectorDBAPI


class WeaviateAPI(VectorDBAPI):
    """Provides a Python interface for interacting with Weaviate,
    facilitating operations such as creating collections, managing
    embeddings, and querying for similar embeddings.

    It only supports cosine similarity for now.
    """

    def __init__(
        self,
        url: str = "http://localhost:8080",
        grpc_url: str = "http://localhost:50051",
        auth_api_key: str = None,
    ) -> None:
        """Initializes the Weaviate API client with connection details.

        @type url: str
        @param url: URL of the Weaviate instance, defaults to
            U{localhost:8080}.
        @type grpc_url: str
        @param grpc_url: URL of the gRPC Weaviate instance, defaults to
            U{localhost:50051}.
        @type auth_api_key: str
        @param auth_api_key: API key for authentication. Defaults to
            C{None}.
        """
        if auth_api_key is not None:
            auth_api_key = weaviate.AuthApiKey(auth_api_key)

        if "localhost" in url:
            self.client = weaviate.connect_to_local()
        else:
            # make sure the gRPC port is correct
            self.client = weaviate.connect_to_custom(
                http_host=url.split("://")[1].split(":")[0],
                http_port=url.split("://")[1].split(":")[1],
                http_secure=url.split("://")[0] == "https",
                grpc_host=grpc_url.split("://")[1].split(":")[0],
                grpc_port=grpc_url.split("://")[1].split(":")[1],
                grpc_secure=grpc_url.split("://")[0] == "https",
                auth_credentials=auth_api_key,
            )

    def create_collection(
        self, collection_name: str, properties: List[str] = None
    ) -> None:
        """Creates a new collection in the Weaviate database.

        @type collection_name: str
        @param collection_name: Name of the collection to create.
        @type properties: List[str]
        @param properties: List of properties for the collection.
            Defaults to None.
        """
        self.collection_name = collection_name
        self.properties = properties

        if not self.client.collections.exists(self.collection_name):
            # print(f"Collection {self.collection_name} does not exist. Creating...")

            properties = []
            for prop in self.properties:
                properties.append(
                    wvc.Property(
                        name=prop,
                        data_type=wvc.DataType.TEXT,
                        skip_vectorization=True,
                    )
                )

            self.collection = self.client.collections.create(
                name=self.collection_name,
                properties=properties,
                vector_index_config=wvc.Configure.VectorIndex.hnsw(
                    distance_metric=wvc.VectorDistance.COSINE
                ),
            )
            print(f"Collection {self.collection_name} created.")
        else:
            self.collection = self.client.collections.get(self.collection_name)
            print(f"Collection {self.collection_name} already exists.")

    def delete_collection(self) -> None:
        """Deletes a collection from the Weaviate database."""
        try:
            self.client.collections.delete(self.collection_name)
            print(f"Collection {self.collection_name} deleted.")
        except Exception:
            print(f"Collection {self.collection_name} does not exist.")

    def insert_embeddings(
        self,
        uuids: List[str],
        embeddings: List[List[float]],
        payloads: List[Dict[str, Any]],
        batch_size: int = 100,
    ) -> None:
        """Inserts embeddings with associated payloads into a
        collection.

        @type uuids: List[str]
        @param uuids: List of UUIDs for the embeddings.
        @type embeddings: List[List[float]]
        @param embeddings: List of embeddings.
        @type payloads: List[Dict[str, Any]]
        @param payloads: List of payloads.
        @type batch_size: int
        @param batch_size: Batch size for inserting the embeddings.
        """
        data = []
        for i, embedding in enumerate(embeddings):
            data.append(
                wvc.DataObject(
                    properties=payloads[i], uuid=uuids[i], vector=embedding
                )
            )

            if len(data) == batch_size:
                self.collection.data.insert_many(data)
                data = []

        if len(data) > 0:
            self.collection.data.insert_many(data)

    def search_similar_embeddings(
        self, embedding: List[float], top_k: int = 10
    ) -> Tuple[List[str], List[float]]:
        """Searches for embeddings similar to a given vector.

        @type embedding: List[float]
        @param embedding: Embedding to find similar embeddings for.
        @type top_k: int
        @param top_k: Number of similar embeddings to find.
        @rtype uuids: List[str]
        @return uuids: List of UUIDs of the similar embeddings.
        @rtype scores: List[float]
        @return scores: List of similarity scores.
        """
        response = self.collection.query.near_vector(
            near_vector=embedding,
            limit=top_k,
            return_metadata=wvc.query.MetadataQuery(distance=True),
        )

        uuids = []
        scores = []
        for result in response.objects:
            uuids.append(str(result.uuid))
            scores.append(1 - result.metadata.distance)

        return uuids, scores

    def get_similarity_scores(
        self,
        reference_id: str,
        other_ids: List[str],
        sort_distances: bool = False,
    ) -> Tuple[List[str], List[float]]:
        """Calculates the similarity score between the reference
        embedding and the specified embeddings.

        @type reference_id: str
        @param reference_id: UUID of the reference embedding.
        @type other_ids: List[str]
        @param other_ids: List of UUIDs of the embeddings to compare to.
        @type sort_distances: bool
        @param sort_distances: Whether to sort the results by distance
            or keep order of the UUIDs. Defaults to False.
        @rtype ids: List[str]
        @return ids: List of UUIDs of the embeddings.
        @rtype scores: List[float]
        @return scores: List of similarity scores.
        """
        response = self.collection.query.near_object(
            near_object=reference_id,
            limit=len(other_ids),
            filters=FilterMetadata.ById.contains_any(other_ids),
            return_metadata=wvc.query.MetadataQuery(distance=True),
        )
        # near object gives order by distance result

        if sort_distances:
            ids, scores = [], []
            for result in response.objects:
                ids.append(str(result.uuid))
                scores.append(1 - result.metadata.distance)
        else:
            ids = other_ids
            scores = [0] * len(other_ids)
            for result in response.objects:
                scores[other_ids.index(str(result.uuid))] = (
                    1 - result.metadata.distance
                )

        return ids, scores

    def compute_similarity_matrix(self) -> List[List[float]]:
        """Calculates the similarity matrix for all the embeddings in
        the collection. @note: This is a very inefficient
        implementation. For large numbers of embeddings, calculate the
        similarity matrix by hand
        (sklearn.metrics.pairwise.cosine_similarity).

        @rtype sim_matrix: List[List[float]]
        @return sim_matrix: Similarity matrix for all the embeddings in
            the collection.
        """
        uuids = self.retrieve_all_ids()

        sim_matrix = []
        for uuid in uuids:
            ids, scores = self.get_similarity_scores(
                uuid, uuids, sort_distances=False
            )
            sim_matrix.append(scores)

        return sim_matrix

    def retrieve_embeddings_by_ids(
        self, uuids: List[str]
    ) -> List[List[float]]:
        """Gets the embeddings for the specified UUIDs, up to a maximum
        of 10000.

        @type uuids: List[str]
        @param uuids: List of UUIDs of the embeddings to get.
        @rtype embeddings: List[List[float]]
        @return embeddings: List of embeddings.
        """
        embeddings = []

        response = self.collection.query.fetch_objects(
            limit=10000,
            filters=FilterMetadata.ById.contains_any(uuids),
            include_vector=True,
        )

        # Create a dictionary mapping UUIDs to embeddings
        uuid_embedding_map = {
            str(result.uuid): result.vector for result in response.objects
        }
        # Retrieve embeddings in the order of the provided UUIDs
        embeddings = [
            uuid_embedding_map[uuid]
            for uuid in uuids
            if uuid in uuid_embedding_map
        ]

        return embeddings

    def retrieve_payloads_by_ids(
        self, uuids: List[str], properties: Optional[List[str]] = None
    ) -> List[Dict[str, Any]]:
        """Gets the payloads for the specified UUIDs, up to a maximum of
        10000.

        @type uuids: List[str]
        @param uuids: List of UUIDs of the embeddings to get.
        @type properties: List[str]
        @param properties: List of properties to retrieve.
        @rtype payloads: List[Dict[str, Any]]
        @return payloads: List of payloads.
        """
        payloads = []

        response = self.collection.query.fetch_objects(
            limit=10000,
            filters=FilterMetadata.ById.contains_any(uuids),
            return_properties=properties,
        )

        # Create a dictionary mapping UUIDs to payloads
        uuid_payload_map = {
            str(result.uuid): result.properties for result in response.objects
        }
        # Retrieve payloads in the order of the provided UUIDs
        payloads = [
            uuid_payload_map[uuid]
            for uuid in uuids
            if uuid in uuid_payload_map
        ]

        return payloads

    def retrieve_all_ids(self) -> List[str]:
        """Gets all the UUIDs in the Weaviate collection.

        @rtype uuids: List[str]
        @return uuids: List of UUIDs.
        """
        uuids = []
        for item in self.collection.iterator():
            uuids.append(str(item.uuid))

        return uuids

    def retrieve_all_embeddings(self) -> Tuple[List[str], List[List[float]]]:
        """Gets all the embeddings and UUIDs in the Weaviate collection.

        @rtype uuids: List[str]
        @return uuids: List of UUIDs.
        @rtype embeddings: List[List[float]]
        @return embeddings: List of embeddings.
        """
        uuids = []
        embeddings = []

        for item in self.collection.iterator(include_vector=True):
            uuids.append(str(item.uuid))
            embeddings.append(item.vector)

        return uuids, embeddings



from typing import Final, List

import pytest

from luxonis_ml.data import LabelType, LuxonisLoader, LuxonisParser
from luxonis_ml.enums import DatasetType
from luxonis_ml.utils import environ

URL_PREFIX: Final[str] = "gs://luxonis-test-bucket/luxonis-ml-test-data"
WORK_DIR: Final[str] = "tests/data/parser_datasets"


@pytest.fixture(scope="module", autouse=True)
def prepare_dir():
    import os
    import shutil

    os.makedirs(WORK_DIR, exist_ok=True)
    yield
    shutil.rmtree(WORK_DIR)


@pytest.mark.parametrize(
    ("dataset_type", "url", "expected_label_types"),
    [
        (
            DatasetType.COCO,
            "COCO_people_subset.zip",
            [
                LabelType.BOUNDINGBOX,
                LabelType.KEYPOINTS,
                LabelType.SEGMENTATION,
                LabelType.CLASSIFICATION,
            ],
        ),
        (
            DatasetType.COCO,
            "Thermal_Dogs_and_People.v1-resize-416x416.coco.zip",
            [LabelType.BOUNDINGBOX, LabelType.CLASSIFICATION],
        ),
        (
            DatasetType.VOC,
            "Thermal_Dogs_and_People.v1-resize-416x416.voc.zip",
            [LabelType.BOUNDINGBOX, LabelType.CLASSIFICATION],
        ),
        (
            DatasetType.DARKNET,
            "Thermal_Dogs_and_People.v1-resize-416x416.darknet.zip",
            [LabelType.BOUNDINGBOX, LabelType.CLASSIFICATION],
        ),
        (
            DatasetType.YOLOV4,
            "Thermal_Dogs_and_People.v1-resize-416x416.yolov4pytorch.zip",
            [LabelType.BOUNDINGBOX, LabelType.CLASSIFICATION],
        ),
        (
            DatasetType.YOLOV6,
            "Thermal_Dogs_and_People.v1-resize-416x416.mt-yolov6.zip",
            [LabelType.BOUNDINGBOX, LabelType.CLASSIFICATION],
        ),
        (
            DatasetType.CREATEML,
            "Thermal_Dogs_and_People.v1-resize-416x416.createml.zip",
            [LabelType.BOUNDINGBOX, LabelType.CLASSIFICATION],
        ),
        (
            DatasetType.TFCSV,
            "Thermal_Dogs_and_People.v1-resize-416x416.tensorflow.zip",
            [LabelType.BOUNDINGBOX, LabelType.CLASSIFICATION],
        ),
        (
            DatasetType.SEGMASK,
            "D2_Tile.png-mask-semantic.zip",
            [LabelType.SEGMENTATION, LabelType.CLASSIFICATION],
        ),
        (
            DatasetType.CLSDIR,
            "Flowers_Classification.v2-raw.folder.zip",
            [LabelType.CLASSIFICATION],
        ),
        (
            DatasetType.SOLO,
            "D1_ParkingSlot-solo.zip",
            [LabelType.BOUNDINGBOX, LabelType.SEGMENTATION],
        ),
        (
            DatasetType.COCO,
            "roboflow://team-roboflow/coco-128/2/coco",
            [LabelType.BOUNDINGBOX, LabelType.CLASSIFICATION],
        ),
    ],
)
def test_dir_parser(
    dataset_type: DatasetType, url: str, expected_label_types: List[LabelType]
):
    if not url.startswith("roboflow://"):
        url = f"{URL_PREFIX}/{url}"

    elif environ.ROBOFLOW_API_KEY is None:
        pytest.skip("Roboflow API key is not set")

    parser = LuxonisParser(
        url,
        dataset_name=f"test-{dataset_type}",
        delete_existing=True,
        save_dir=WORK_DIR,
    )
    dataset = parser.parse()
    assert len(dataset) > 0
    loader = LuxonisLoader(dataset)
    _, ann = next(iter(loader))
    label_types = {label_type for _, label_type in ann.values()}
    assert label_types == set(expected_label_types)
    dataset.delete_dataset()


def test_custom_tasks():
    parser = LuxonisParser(
        f"{URL_PREFIX}/Thermal_Dogs_and_People.v1-resize-416x416.coco.zip",
        dataset_name="test-custom-tasks",
        delete_existing=True,
        save_dir=WORK_DIR,
        task_mapping={LabelType.BOUNDINGBOX: "object_detection"},
    )
    dataset = parser.parse()
    assert len(dataset) > 0
    tasks = dataset.get_tasks()
    assert set(tasks) == {"object_detection", "classification"}



from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import polars as pl

from luxonis_ml.data import DatasetIterator

from .base_parser import BaseParser, ParserOutput


class TensorflowCSVParser(BaseParser):
    """Parses directory with TensorflowCSV annotations to LDF.

    Expected format::

        dataset_dir/
        ├── train/
        │   ├── img1.jpg
        │   ├── img2.jpg
        │   ├── ...
        │   └── _annotations.csv
        ├── valid/
        └── test/

    This is one of the formats that can be generated by
    U{Roboflow <https://roboflow.com/>}.
    """

    @staticmethod
    def validate_split(split_path: Path) -> Optional[Dict[str, Any]]:
        if not split_path.exists():
            return None
        if not BaseParser._list_images(split_path):
            return None
        if not (split_path / "_annotations.csv").exists():
            return None
        return {
            "image_dir": split_path,
            "annotation_path": split_path / "_annotations.csv",
        }

    @staticmethod
    def validate(dataset_dir: Path) -> bool:
        for split in ["train", "valid", "test"]:
            split_path = dataset_dir / split
            if TensorflowCSVParser.validate_split(split_path) is None:
                return False
        return True

    def from_dir(
        self, dataset_dir: Path
    ) -> Tuple[List[str], List[str], List[str]]:
        added_train_imgs = self._parse_split(
            image_dir=dataset_dir / "train",
            annotation_path=dataset_dir / "train" / "_annotations.csv",
        )
        added_val_imgs = self._parse_split(
            image_dir=dataset_dir / "valid",
            annotation_path=dataset_dir / "valid" / "_annotations.csv",
        )
        added_test_imgs = self._parse_split(
            image_dir=dataset_dir / "test",
            annotation_path=dataset_dir / "test" / "_annotations.csv",
        )
        return added_train_imgs, added_val_imgs, added_test_imgs

    def from_split(
        self, image_dir: Path, annotation_path: Path
    ) -> ParserOutput:
        """Parses annotations from TensorflowCSV format to LDF.
        Annotations include classification and object detection.

        @type image_dir: Path
        @param image_dir: Path to directory with images
        @type annotation_path: Path
        @param annotation_path: Path to annotation CSV file
        @rtype: L{ParserOutput}
        @return: Annotation generator, list of classes names, skeleton
            dictionary for
        """
        df = pl.read_csv(annotation_path).filter(
            pl.col("filename").is_not_null()
        )
        images_annotations = {}

        class_names = set(df["class"])
        for row in df.rows(named=True):
            path = str(image_dir / str(row["filename"]))
            if path not in images_annotations:
                images_annotations[path] = {
                    "classes": [],
                    "bboxes": [],
                }

            class_name = row["class"]
            images_annotations[path]["classes"].append(class_name)

            height = row["height"]
            width = row["width"]
            xmin = row["xmin"]
            ymin = row["ymin"]
            xmax = row["xmax"]
            ymax = row["ymax"]
            bbox_xywh = np.array(
                [xmin, ymin, xmax - xmin, ymax - ymin], dtype=float
            )
            bbox_xywh[::2] /= width
            bbox_xywh[1::2] /= height
            bbox_xywh = bbox_xywh.tolist()
            images_annotations[path]["bboxes"].append((class_name, bbox_xywh))

        def generator() -> DatasetIterator:
            for path in images_annotations:
                curr_annotations = images_annotations[path]
                for class_name in curr_annotations["classes"]:
                    yield {
                        "file": path,
                        "annotation": {
                            "type": "classification",
                            "class": class_name,
                        },
                    }
                for bbox_class, (x, y, w, h) in curr_annotations["bboxes"]:
                    yield {
                        "file": path,
                        "annotation": {
                            "type": "boundingbox",
                            "class": bbox_class,
                            "x": x,
                            "y": y,
                            "w": w,
                            "h": h,
                        },
                    }

        added_images = self._get_added_images(generator())

        return generator(), list(class_names), {}, added_images



from pydantic import BaseModel, Field

from ..enums import DataType


class Metadata(BaseModel):
    """Represents metadata of a model.

    @type name: str
    @ivar name: Name of the model.
    @type path: str
    @ivar path: Relative path to the model executable.
    """

    name: str = Field(description="Name of the model.")
    path: str = Field(description="Relative path to the model executable.")
    precision: DataType = Field(
        DataType.FLOAT32, description="Precision of the model weights."
    )



import rich.progress


def make_progress_bar() -> rich.progress.Progress:
    return rich.progress.Progress(
        rich.progress.TextColumn("[progress.description]{task.description}"),
        rich.progress.BarColumn(),
        rich.progress.TaskProgressColumn(),
        rich.progress.MofNCompleteColumn(),
        rich.progress.TimeRemainingColumn(),
    )



import pydantic
import pytest

from luxonis_ml.data.datasets.annotation import (
    BBoxAnnotation,
    KeypointAnnotation,
    PolylineSegmentationAnnotation,
)


def test_bbox_no_auto_clip():
    base_dict = {"x": 0, "y": 0, "w": 0, "h": 0}
    for k in ["x", "y", "w", "h"]:
        for v in [-2.1, 2.3, -3.3, 3]:
            with pytest.raises(pydantic.ValidationError):
                curr_dict = base_dict.copy()
                curr_dict[k] = v
                BBoxAnnotation(**curr_dict)


def test_bbox_auto_clip():
    base_dict = {"x": 0, "y": 0, "w": 0, "h": 0}
    for k in ["x", "y", "w", "h"]:
        for v in [-1.1, 1.3, -1.3, 2]:
            curr_dict = base_dict.copy()
            curr_dict[k] = v
            bbox_ann = BBoxAnnotation(**curr_dict)
            assert 0 <= bbox_ann.x <= 1
            assert 0 <= bbox_ann.y <= 1
            assert 0 <= bbox_ann.w <= 1
            assert 0 <= bbox_ann.h <= 1


def test_bbox_clip_sum():
    bbox_ann = BBoxAnnotation(**{"x": 0.9, "y": 0, "w": 0.2, "h": 0})
    assert bbox_ann.x + bbox_ann.w <= 1
    bbox_ann = BBoxAnnotation(**{"x": 1.2, "y": 0, "w": 0.2, "h": 0})
    assert bbox_ann.x + bbox_ann.w <= 1
    bbox_ann = BBoxAnnotation(**{"x": 0, "y": 0.9, "w": 0, "h": 0.2})
    assert bbox_ann.y + bbox_ann.h <= 1
    bbox_ann = BBoxAnnotation(**{"x": 0, "y": 1.2, "w": 0, "h": 0.2})
    assert bbox_ann.y + bbox_ann.h <= 1


def test_kpt_no_auto_clip():
    with pytest.raises(pydantic.ValidationError):
        KeypointAnnotation(**{"keypoints": [(-2.1, 1.1, 0)]})
    with pytest.raises(pydantic.ValidationError):
        KeypointAnnotation(**{"keypoints": [(0.1, 2.1, 1)]})
    with pytest.raises(pydantic.ValidationError):
        KeypointAnnotation(**{"keypoints": [(0.1, 1.1, 2), (0.1, 2.1, 1)]})


def test_kpt_auto_clip():
    kpt_ann = KeypointAnnotation(**{"keypoints": [(-1.1, 1.1, 0)]})
    assert (
        0 <= kpt_ann.keypoints[0][0] <= 1 and 0 <= kpt_ann.keypoints[0][1] <= 1
    )
    kpt_ann = KeypointAnnotation(**{"keypoints": [(0.1, 1.1, 1)]})
    assert (
        0 <= kpt_ann.keypoints[0][0] <= 1 and 0 <= kpt_ann.keypoints[0][1] <= 1
    )
    kpt_ann = KeypointAnnotation(**{"keypoints": [(-2, 2, 2)]})
    assert (
        0 <= kpt_ann.keypoints[0][0] <= 1 and 0 <= kpt_ann.keypoints[0][1] <= 1
    )


def test_poly_no_auto_clip():
    with pytest.raises(pydantic.ValidationError):
        PolylineSegmentationAnnotation(
            **{"points": [(-2.1, 1.1), (-2.1, 2.1), (-0.1, -2.1)]}
        )


def test_poly_auto_clip():
    poly_ann = PolylineSegmentationAnnotation(
        **{"points": [(-0.1, 1.1), (-2, 2), (-0.1, -1.1)]}
    )
    assert 0 <= poly_ann.points[0][0] <= 1 and 0 <= poly_ann.points[0][1] <= 1
    assert 0 <= poly_ann.points[1][0] <= 1 and 0 <= poly_ann.points[1][1] <= 1
    assert 0 <= poly_ann.points[2][0] <= 1 and 0 <= poly_ann.points[2][1] <= 1



import shutil
from collections import defaultdict
from pathlib import Path
from typing import Dict

import cv2
import numpy as np
import pytest

from luxonis_ml.data import BucketStorage, LuxonisDataset, LuxonisLoader

DATASET_NAME = "test-task-ingestion"
DATA_DIR = Path("tests/data/test_task_ingestion")
STEP = 10


@pytest.fixture(autouse=True, scope="module")
def prepare_dir():
    DATA_DIR.mkdir(parents=True, exist_ok=True)

    yield

    shutil.rmtree(DATA_DIR)


def make_image(i) -> Path:
    path = DATA_DIR / f"img_{i}.jpg"
    if not path.exists():
        img = np.zeros((512, 512, 3), dtype=np.uint8)
        img[0:10, 0:10] = np.random.randint(
            0, 255, (10, 10, 3), dtype=np.uint8
        )
        cv2.imwrite(str(path), img)
    return path


def compute_histogram(dataset: LuxonisDataset) -> Dict[str, int]:
    classes = defaultdict(int)
    loader = LuxonisLoader(dataset, force_resync=True)
    for _, record in loader:
        for task, _ in record.items():
            classes[task] += 1

    return dict(classes)


@pytest.mark.parametrize(
    ("bucket_storage",),
    [
        (BucketStorage.LOCAL,),
        (BucketStorage.S3,),
        (BucketStorage.GCS,),
    ],
)
def test_task_ingestion(
    bucket_storage: BucketStorage, platform_name: str, python_version: str
):
    dataset = LuxonisDataset(
        f"{DATASET_NAME}-{bucket_storage.value}-{platform_name}-{python_version}",
        bucket_storage=bucket_storage,
        delete_existing=True,
        delete_remote=True,
    )

    def generator1():
        for i in range(STEP):
            path = make_image(i)
            yield {
                "file": str(path),
                "annotation": {
                    "type": "boundingbox",
                    "class": "dog",
                    "task": "animals-boxes",
                    "x": 0.1,
                    "y": 0.1,
                    "w": 0.1,
                    "h": 0.1,
                },
            }
            yield {
                "file": str(path),
                "annotation": {
                    "type": "boundingbox",
                    "class": "cat",
                    "task": "animals-boxes",
                    "x": 0.5,
                    "y": 0.5,
                    "w": 0.1,
                    "h": 0.3,
                },
            }
            yield {
                "file": str(path),
                "annotation": {
                    "type": "polyline",
                    "class": "water",
                    "task": "land-segmentation",
                    "points": [
                        (0.1, 0.1),
                        (0.2, 0.8),
                        (0.8, 0.3),
                        (0, 0.5),
                        (0.5, 0.5),
                    ],
                },
            }
            yield {
                "file": str(path),
                "annotation": {
                    "type": "polyline",
                    "class": "grass",
                    "task": "land-segmentation",
                    "points": [(0.1, 0.5), (0.6, 0.6), (0.7, 0.7)],
                },
            }

    dataset.add(generator1()).make_splits((1, 0, 0))

    classes_list, classes = dataset.get_classes()

    assert set(classes_list) == {"dog", "cat", "water", "grass"}
    assert set(classes["land-segmentation"]) == {"water", "grass"}
    assert set(classes["animals-boxes"]) == {"dog", "cat"}

    assert compute_histogram(dataset) == {
        "animals-boxes": STEP,
        "land-segmentation": STEP,
    }

    def generator2():
        for i in range(STEP, 2 * STEP):
            path = make_image(i)
            yield {
                "file": str(path),
                "annotation": {
                    "type": "boundingbox",
                    "class": "dog",
                    "x": 0.1,
                    "y": 0.1,
                    "w": 0.1,
                    "h": 0.1,
                },
            }
            yield {
                "file": str(path),
                "annotation": {
                    "type": "boundingbox",
                    "class": "cat",
                    "x": 0.5,
                    "y": 0.5,
                    "w": 0.1,
                    "h": 0.3,
                },
            }

    dataset.add(generator2()).make_splits((1, 0, 0))
    classes_list, classes = dataset.get_classes()

    assert set(classes_list) == {"dog", "cat", "water", "grass"}
    assert set(classes["land-segmentation"]) == {"water", "grass"}
    assert set(classes["animals-boxes"]) == {"dog", "cat"}

    assert compute_histogram(dataset) == {
        "animals-boxes": 2 * STEP,
        "land-segmentation": STEP,
    }

    def generator3():
        for i in range(2 * STEP, 3 * STEP):
            path = make_image(i)
            yield {
                "file": str(path),
                "annotation": {
                    "type": "boundingbox",
                    "task": "animals-boxes",
                    "class": "dog",
                    "x": 0.15,
                    "y": 0.25,
                    "w": 0.1,
                    "h": 0.1,
                },
            }
            yield {
                "file": str(path),
                "annotation": {
                    "type": "polyline",
                    "class": "water",
                    "points": [
                        (0.1, 0.7),
                        (0.5, 0.2),
                        (0.3, 0.3),
                        (0.12, 0.45),
                    ],
                },
            }

    dataset.add(generator3()).make_splits((1, 0, 0))
    classes_list, classes = dataset.get_classes()

    assert set(classes_list) == {"dog", "cat", "water", "grass"}
    assert set(classes["land-segmentation"]) == {"water", "grass"}
    assert set(classes["animals-boxes"]) == {"dog", "cat"}

    assert compute_histogram(dataset) == {
        "animals-boxes": 3 * STEP,
        "land-segmentation": 2 * STEP,
    }

    def generator4():
        for i in range(3 * STEP, 4 * STEP):
            path = make_image(i)
            yield {
                "file": str(path),
                "annotation": {
                    "type": "boundingbox",
                    "class": "bike",
                    "x": 0.9,
                    "y": 0.8,
                    "w": 0.1,
                    "h": 0.4,
                },
            }
            yield {
                "file": str(path),
                "annotation": {
                    "type": "polyline",
                    "class": "body",
                    "points": [(0.1, 0.1), (0.7, 0.5), (0.3, 0.3), (0.5, 0.5)],
                },
            }
            yield {
                "file": str(path),
                "annotation": {
                    "type": "polyline",
                    "class": "water",
                    "task": "land-segmentation-2",
                    "points": [(0.1, 0.1), (0.8, 0.2), (0.8, 0.9), (0.1, 0.9)],
                },
            }

    dataset.add(generator4()).make_splits((1, 0, 0))
    classes_list, classes = dataset.get_classes()

    assert set(classes_list) == {
        "dog",
        "cat",
        "water",
        "grass",
        "bike",
        "body",
    }
    assert set(classes["land-segmentation"]) == {"water", "grass"}
    assert set(classes["animals-boxes"]) == {"dog", "cat"}
    assert set(classes["land-segmentation-2"]) == {"water"}
    assert set(classes["segmentation"]) == {"body"}

    assert compute_histogram(dataset) == {
        "animals-boxes": 3 * STEP,
        "land-segmentation": 2 * STEP,
        "land-segmentation-2": STEP,
        "segmentation": STEP,
        "boundingbox": STEP,
    }



from pathlib import Path
from typing import Literal, Optional

from pydantic import model_serializer
from pydantic_settings import BaseSettings, SettingsConfigDict

__all__ = ["Environ", "environ"]


class Environ(BaseSettings):
    """A L{BaseSettings} subclass for storing environment variables."""

    model_config = SettingsConfigDict(
        env_file=".env", env_file_encoding="utf-8", extra="ignore"
    )

    AWS_ACCESS_KEY_ID: Optional[str] = None
    AWS_SECRET_ACCESS_KEY: Optional[str] = None
    AWS_S3_ENDPOINT_URL: Optional[str] = None

    MLFLOW_CLOUDFLARE_ID: Optional[str] = None
    MLFLOW_CLOUDFLARE_SECRET: Optional[str] = None
    MLFLOW_S3_BUCKET: Optional[str] = None
    MLFLOW_S3_ENDPOINT_URL: Optional[str] = None
    MLFLOW_TRACKING_URI: Optional[str] = None

    POSTGRES_USER: Optional[str] = None
    POSTGRES_PASSWORD: Optional[str] = None
    POSTGRES_HOST: Optional[str] = None
    POSTGRES_PORT: Optional[str] = None
    POSTGRES_DB: Optional[str] = None

    LUXONISML_BUCKET: Optional[str] = None
    LUXONISML_BASE_PATH: Path = Path.home() / "luxonis_ml"
    LUXONISML_TEAM_ID: str = "offline"

    ROBOFLOW_API_KEY: Optional[str] = None

    GOOGLE_APPLICATION_CREDENTIALS: Optional[str] = None

    LOG_LEVEL: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = (
        "INFO"
    )

    @model_serializer(when_used="always", mode="plain", return_type=str)
    def _serialize_environ(self) -> str:
        return "{}"


environ = Environ()



import json
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from PIL import Image

from luxonis_ml.data import DatasetIterator

from .base_parser import BaseParser, ParserOutput


class CreateMLParser(BaseParser):
    """Parses directory with CreateML annotations to LDF.

    Expected format::

        dataset_dir/
        ├── train/
        │   ├── img1.jpg
        │   ├── img2.jpg
        │   └── ...
        │   └── _annotations.createml.json
        ├── valid/
        └── test/

    This is one of the formats that can be generated by
    U{Roboflow <https://roboflow.com/>}.
    """

    @staticmethod
    def validate_split(split_path: Path) -> Optional[Dict[str, Any]]:
        if not split_path.exists():
            return None
        if not (split_path / "_annotations.createml.json").exists():
            return None
        if not BaseParser._list_images(split_path):
            return None
        return {
            "image_dir": split_path,
            "annotation_path": split_path / "_annotations.createml.json",
        }

    @staticmethod
    def validate(dataset_dir: Path) -> bool:
        for split in ["train", "valid", "test"]:
            split_path = dataset_dir / split
            if CreateMLParser.validate_split(split_path) is None:
                return False
        return True

    def from_dir(
        self, dataset_dir: Path
    ) -> Tuple[List[str], List[str], List[str]]:
        added_train_imgs = self._parse_split(
            image_dir=dataset_dir / "train",
            annotation_path=dataset_dir
            / "train"
            / "_annotations.createml.json",
        )
        added_val_imgs = self._parse_split(
            image_dir=dataset_dir / "valid",
            annotation_path=dataset_dir
            / "valid"
            / "_annotations.createml.json",
        )
        added_test_imgs = self._parse_split(
            image_dir=dataset_dir / "test",
            annotation_path=dataset_dir
            / "test"
            / "_annotations.createml.json",
        )

        return added_train_imgs, added_val_imgs, added_test_imgs

    def from_split(
        self, image_dir: Path, annotation_path: Path
    ) -> ParserOutput:
        """Parses annotations from CreateML format to LDF. Annotations
        include classification and object detection.

        @type image_dir: Path
        @param image_dir: Path to directory with images
        @type annotation_path: Path
        @param annotation_path: Path to annotation json file
        @rtype: L{ParserOutput}
        @return: Annotation generator, list of classes names, skeleton
            dictionary for keypoints and list of added images.
        """
        with open(annotation_path) as f:
            annotations_data = json.load(f)

        class_names = set()
        images_annotations = []
        for annotations in annotations_data:
            path = image_dir.absolute().resolve() / annotations["image"]
            if not path.exists():
                continue
            file = str(path)
            img = Image.open(file)
            width, height = img.size

            curr_annotations = {"path": str(path), "classes": [], "bboxes": []}
            for curr_ann in annotations["annotations"]:
                class_name = curr_ann["label"]
                curr_annotations["classes"].append(class_name)
                class_names.add(class_name)

                bbox_ann = curr_ann["coordinates"]
                bbox_xywh = [
                    (bbox_ann["x"] - bbox_ann["width"] / 2) / width,
                    (bbox_ann["y"] - bbox_ann["height"] / 2) / height,
                    bbox_ann["width"] / width,
                    bbox_ann["height"] / height,
                ]
                curr_annotations["bboxes"].append((class_name, bbox_xywh))
            images_annotations.append(curr_annotations)

        def generator() -> DatasetIterator:
            for curr_annotations in images_annotations:
                path = curr_annotations["path"]
                for class_name in curr_annotations["classes"]:
                    yield {
                        "file": path,
                        "annotation": {
                            "type": "classification",
                            "class": class_name,
                        },
                    }
                for bbox_class, (x, y, w, h) in curr_annotations["bboxes"]:
                    yield {
                        "file": path,
                        "annotation": {
                            "type": "boundingbox",
                            "class": bbox_class,
                            "x": x,
                            "y": y,
                            "w": w,
                            "h": h,
                        },
                    }

        added_images = self._get_added_images(generator())

        return generator(), list(class_names), {}, added_images



from typing import Any, Dict, List


def batch2list(data: Dict[str, List]) -> List[Dict[str, Any]]:
    """Convert from a batched target dict to list of normal target
    dicts."""
    if "image_batch" not in data:
        raise ValueError(
            "Batch-based transform should have `image_batch` target"
        )
    batch_size = len(data["image_batch"])
    items = []
    for i in range(batch_size):
        item = {}
        for k, v in data.items():
            if k.endswith("_batch"):
                # ex. {"image_batch": image_batch} -> {"image": image_batch[i]}
                item_k = to_unbatched_name(k)
                item[item_k] = v[i]
            else:
                raise ValueError(
                    f"All key must have '_batch' suffix, got `{k}`"
                )
        items.append(item)
    return items


def list2batch(data: List[Dict[str, Any]]) -> Dict[str, List]:
    """Convert from a list of normal target dicts to a batched target
    dict."""

    if len(data) == 0:
        raise ValueError("The input should have at least one item.")

    item = data[0]
    batch: Dict[str, Any] = {f"{k}_batch": [] for k in item.keys()}
    for item in data:
        for k, v in item.items():
            batch_k = to_batched_name(k)
            batch[batch_k].append(v)

    return batch


def to_unbatched_name(batched_name: str) -> str:
    """Get a normal target name from a batched target name If the given
    name does not have "_batched" suffix, ValueError will be raised."""
    if not batched_name.endswith("_batch"):
        raise ValueError(
            f"Batched target name must have '_batch' suffix, got `{batched_name}`"
        )
    return batched_name.replace("_batch", "")


def to_batched_name(name: str) -> str:
    """Get a unbatched target name from a normal target name If the
    given name already has had "_batched" suffix, ValueError will be
    raised."""

    if name.endswith("_batch"):
        raise ValueError(
            f"Non batched target name must not have '_batch' suffix, got `{name}`"
        )
    return f"{name}_batch"


def concat_batches(batches: List[Dict[str, List]]) -> Dict[str, List]:
    """Concatenate batched targets."""

    n_batches = len(batches)
    if n_batches == 0:
        raise ValueError("The input should have at least one item.")

    keys = list(batches[0].keys())
    out_batch: Dict[str, List] = {k: [] for k in keys}
    for batch in batches:
        for k in keys:
            for item in batch[k]:
                out_batch[k].append(item)
    return out_batch



from abc import ABCMeta
from typing import Callable, Dict, Generic, Optional, Tuple, TypeVar, Union

T = TypeVar("T")


class Registry(Generic[T]):
    def __init__(self, name: str):
        """A Registry class to store and retrieve modules.

        @type name: str
        @param name: Name of the registry
        """
        self._module_dict: Dict[str, T] = {}
        self._name = name

    def __str__(self):
        return f"Registry('{self.name}')"

    def __repr__(self):
        return self.__str__()

    def __len__(self):
        return len(self._module_dict)

    @property
    def name(self):
        return self._name

    @property
    def module_dict(self):
        return self._module_dict

    def get(self, key: str) -> T:
        """Retrieves the registry record for the key.

        @type key: str
        @param key: Name of the registered item, I{e.g.} the class name
            in string format.
        @rtype: type
        @return: Corresponding class if L{key} exists
        @raise KeyError: If L{key} is not in the registry
        """
        module_cls = self._module_dict.get(key, None)
        if module_cls is None:
            raise KeyError(f"Class `{key}` not in the `{self.name}` registry.")
        else:
            return module_cls

    def register_module(
        self,
        name: Optional[str] = None,
        module: Optional[type] = None,
        force: bool = False,
    ) -> Union[type, Callable[[type], type]]:
        """Registers a module.

        Can be used as a decorator or as a normal method:

            >>> registry = Registry(name="modules")
            >>> @registry.register_module()
            ... class Foo:
            ...     pass
            >>> registry.get("Foo")
            <class '__main__.Foo'>
            >>> class Bar:
            ...     pass
            >>> registry.register_module(module=Bar)
            >>> registry.get("Bar")
            <class '__main__.Bar'>

        @type name: Optional[str]
        @param name: Name of the module. If C{None}, then use class name.
            Defaults to None.

        @type module: Optional[type]
        @param module: Module class to be registered. Defaults to None.

        @type force: bool
        @param force: Whether to override an existing class with the same name.
            Defaults to False.

        @rtype: Union[type, Callable[[type], type]]
        @return: Module class or register function if used as a decorator

        @raise KeyError: Raised if class name already exists and C{force==False}
        """

        # use it as a normal method: x.register_module(module=SomeClass)
        if module is not None:
            self._register_module(module=module, module_name=name, force=force)
            return module

        # use it as a decorator: @x.register_module()
        def _register(module: type) -> type:
            self._register_module(module=module, module_name=name, force=force)
            return module

        return _register

    def _register_module(
        self,
        module: type,
        module_name: Optional[str] = None,
        force: bool = False,
    ) -> None:
        """Registers a module by creating a (key, value) pair."""
        if module_name is None:
            module_name = module.__name__

        if not force and module_name in self._module_dict:
            existed_module = self._module_dict[module_name]
            raise KeyError(
                f"`{module_name}` already registred in `{self.name}` registry at `{existed_module.__module__}`."
            )

        self._module_dict[module_name] = module


class AutoRegisterMeta(ABCMeta):
    """Metaclass for automatically registering modules.

    Can be set as a metaclass for abstract base classes. Then, all subclasses will be
    automatically registered under the name of the subclass.

    Example:

        >>> REGISTRY = Registry(name="modules")
        >>> class BaseClass(metaclass=AutoRegisterMeta, registry=REGISTRY):
        ...     pass
        >>> class SubClass(BaseClass):
        ...     pass
        >>> REGISTRY.get("SubClass")
        <class '__main__.SubClass'>
        >>> BaseClass.REGISTRY.get("SubClass")
        <class '__main__.SubClass'>
    """

    REGISTRY: Registry

    def __new__(
        cls,
        name: str,
        bases: Tuple[type, ...],
        attrs: Dict[str, type],
        register: bool = True,
        register_name: Optional[str] = None,
        registry: Optional[Registry] = None,
    ):
        """Automatically register the class.

        @type name: str
        @param name: Class name

        @type bases: Tuple[type, ...]
        @param bases: Base classes

        @type attrs: Dict[str, type]
        @param attrs: Class attributes

        @type register: bool
        @param register: Weather to register the class. Defaults to True.
            Should be set to False for abstract base classes.

        @type register_name: Optional[str]
        @param register_name: Name used for registration.
            If unset, the class name is used. Defaults to None.

        @type registry: Optional[Registry]
        @param registry: Registry to use for registration.
            Defaults to None. Has to be set in the base class.
        """
        new_class = super().__new__(cls, name, bases, attrs)
        if not hasattr(new_class, "REGISTRY"):
            if registry is not None:
                new_class.REGISTRY = registry
            elif register:
                raise ValueError(
                    "Registry has to be set in the base class or passed as an argument."
                )
        if register:
            registry = registry if registry is not None else new_class.REGISTRY
            registry.register_module(
                name=register_name or name, module=new_class
            )
        return new_class



import builtins
import inspect
import logging
import warnings
from functools import wraps
from typing import Dict, Optional, Type

from rich import print as rprint
from rich.console import Console
from rich.logging import RichHandler
from rich.theme import Theme

from .environ import environ


def setup_logging(
    *,
    file: Optional[str] = None,
    use_rich: bool = False,
    level: Optional[str] = None,
    configure_warnings: bool = True,
    rich_print: bool = False,
    **kwargs,
) -> None:
    """Globally configures logging.

    Configures a standar Luxonis logger. Optionally utilizes rich
    library, configures handling of warnings (from warnings module) and
    saves the logs to a file.

    @type file: str or None
    @param file: Path to a file where logs will be saved. If None, logs
        will not be saved. Defaults to None.
    @type use_rich: bool
    @param use_rich: If True, rich library will be used for logging.
        Defaults to False.
    @type level: str or None
    @param level: Logging level. One of "DEBUG", "INFO", "WARNING",
        "ERROR", and "CRITICAL". Defaults to "INFO". The log level can
        be changed using "LOG_LEVEL" environment variable.
    @type configure_warnings: bool
    @param configure_warnings: If True, warnings will be logged.
        Defaults to True.
    @type rich_print: bool
    @param rich_print: If True, builtins.print will be replaced with
        rich.print. Defaults to False.
    @param kwargs: Additional arguments passed to RichHandler.
    """

    if rich_print:
        builtins.print = rprint

    # NOTE: So we can simply run e.g. `LOG_LEVEL=DEBUG python ...`
    level = level or environ.LOG_LEVEL

    if level not in ("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"):
        raise ValueError(
            f"Invalid logging level: {level}. "
            "Use one of 'DEBUG', 'INFO', 'WARNING', 'ERROR', or 'CRITICAL'."
        )

    handlers = []

    format = file_format = "%(asctime)s [%(levelname)s] %(message)s"
    datefmt = "[%Y-%m-%d %H:%M:%S]"
    if use_rich:
        format = "%(message)s"
        # NOTE: The default rich logging colors are weird.
        theme = Theme(
            {
                "logging.level.debug": "magenta",
                "logging.level.info": "green",
                "logging.level.warning": "yellow",
            }
        )
        console = Console(theme=theme)
        handlers.append(
            RichHandler(
                console=console,
                **{
                    **{
                        "rich_tracebacks": True,
                        "tracebacks_show_locals": False,
                        "omit_repeated_times": False,
                        "show_time": False,
                    },
                    **kwargs,
                },
            )
        )
    else:
        handlers.append(logging.StreamHandler())

    if file is not None:
        file_handler = logging.FileHandler(file)
        file_handler.setFormatter(
            logging.Formatter(file_format, datefmt=datefmt)
        )
        handlers.append(file_handler)

    logging.basicConfig(
        level=level, format=format, datefmt=datefmt, handlers=handlers
    )

    def _custom_warning_handler(
        message: str,
        category: Type[Warning],
        filename: str,
        lineno: int,
        _=None,
        line: Optional[str] = None,
    ):
        logger = logging.getLogger(filename)
        message = warnings.formatwarning(
            message,
            category,
            filename,
            lineno,
            line,
        )
        logger.warning(message)

    if configure_warnings:
        warnings.showwarning = _custom_warning_handler


def reset_logging() -> None:
    """Resets the logging module back to its default state."""
    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)


def deprecated(
    *args: str,
    suggest: Optional[Dict[str, str]] = None,
    additional_message: Optional[str] = None,
    altogether: bool = False,
):
    """Decorator to mark a function or its parameters as deprecated.

    Example:

        >>> @deprecated("old_arg",
        ...             "another_old_arg",
        ...             suggest={"old_arg": "new_arg"},
        ...             additional_message="Usage of 'old_arg' is discouraged.")
        ...def my_func(old_arg, another_old_arg, new_arg=None):
        ...   pass
        >>> my_func("foo")
        >>> # DeprecationWarning: Argument 'old_arg'
        ... # in function `my_func` is deprecated and
        ... # will be removed in future versions.
        ... # Use 'new_arg' instead.
        ... # Usage of 'old_arg' is discouraged.

    @type args: str
    @param args: The names of the deprecated parameters.
    @type suggest: Dict[str, str]
    @param suggest: Suggested replacement parameters.
    @type additional_message: str
    @param additional_message: Additional message to display.
        If provided, it will be appended to the warning message.
    @type altogether: bool
    @param altogether: If True, the whole function is
        marked as deprecated. Defaults to False.
    """

    def decorator(func):
        sig = inspect.signature(func)

        @wraps(func)
        def wrapper(*f_args, **f_kwargs):
            fname = func.__name__
            if altogether:
                msg = f"'{fname}' is deprecated and will be removed in future versions."
                if additional_message:
                    msg += f" {additional_message}"
                warnings.warn(msg, DeprecationWarning, stacklevel=2)

            if args:
                pos_arg_names = list(sig.parameters.keys())[: len(f_args)]

                for arg_name in pos_arg_names:
                    if arg_name in args:
                        _warn_deprecated(
                            arg_name, fname, suggest, additional_message
                        )

                for arg_name in f_kwargs:
                    if arg_name in args:
                        _warn_deprecated(
                            arg_name, fname, suggest, additional_message
                        )

            return func(*f_args, **f_kwargs)

        return wrapper

    return decorator


def _warn_deprecated(arg_name, fname, suggest, additional_message):
    replacement = suggest.get(arg_name) if suggest else None
    msg = (
        f"Argument '{arg_name}' in function '{fname}' "
        "is deprecated and will be removed in future versions."
    )
    if replacement:
        msg += f" Use '{replacement}' instead."
    if additional_message:
        msg += f" {additional_message}"
    warnings.warn(msg, DeprecationWarning, stacklevel=3)



import hashlib
import math
from typing import Dict, Iterator, List, Tuple

import cv2
import numpy as np

from luxonis_ml.data import LabelType
from luxonis_ml.data.loaders import Labels


def _task_to_rgb(string: str) -> tuple:
    h = int(hashlib.md5(string.encode()).hexdigest(), 16)
    r = (h & 0xFF0000) >> 16
    g = (h & 0x00FF00) >> 8
    b = h & 0x0000FF

    return (r, g, b)


def create_text_image(
    text: str,
    width: int,
    height: int,
    font_size: float = 0.7,
    bg_color: Tuple[int, int, int] = (255, 255, 255),
    text_color: Tuple[int, int, int] = (0, 0, 0),
):
    """Creates an image with the given text centered in the image.

    @type text: str
    @param text: The text to display.
    @type width: int
    @param width: The width of the image.
    @type height: int
    @param height: The height of the image.
    @type font_size: float
    @param font_size: The font size of the text. Default is 0.7.
    @type bg_color: Tuple[int, int, int]
    @param bg_color: The background color of the image. Default is
        white.
    @type text_color: Tuple[int, int, int]
    @param text_color: The color of the text. Default is black.
    """
    img = np.full((height, width, 3), bg_color, dtype=np.uint8)

    font = cv2.FONT_HERSHEY_SIMPLEX

    text_size = cv2.getTextSize(text, font, font_size, 1)[0]

    text_x = (width - text_size[0]) // 2
    text_y = (height + text_size[1]) // 2

    cv2.putText(
        img,
        text,
        (text_x, text_y),
        font,
        font_size,
        text_color,
        1,
        cv2.LINE_AA,
    )

    return img


def concat_images(
    image_dict: Dict[str, np.ndarray],
    padding: int = 10,
    label_height: int = 30,
):
    """Concatenates images into a single image with labels.

    It will attempt to create a square grid of images.

    @type image_dict: Dict[str, np.ndarray]
    @param image_dict: A dictionary mapping image names to images.
    @type padding: int
    @param padding: The padding between images. Default is 10.
    @type label_height: int
    @param label_height: The height of the label. Default
    @rtype: np.ndarray
    @return: The concatenated image.
    """
    n_images = len(image_dict)
    n_cols = math.ceil(math.sqrt(n_images))
    n_rows = math.ceil(n_images / n_cols)

    max_h = max(img.shape[0] for img in image_dict.values())
    max_w = max(img.shape[1] for img in image_dict.values())

    cell_height = max_h + 2 * padding + label_height
    cell_width = max_w + 2 * padding

    output = np.full(
        (cell_height * n_rows, cell_width * n_cols, 3), 255, dtype=np.uint8
    )

    for idx, (name, img) in enumerate(image_dict.items()):
        i = idx // n_cols
        j = idx % n_cols

        y_start = i * cell_height
        x_start = j * cell_width

        label = create_text_image(name, cell_width, label_height)
        output[
            y_start : y_start + label_height, x_start : x_start + cell_width
        ] = label

        h, w = img.shape[:2]
        y_img = y_start + label_height + padding
        x_img = x_start + padding
        output[y_img : y_img + h, x_img : x_img + w] = img

    return output


def _label_type_iterator(
    labels: Labels, label_type: LabelType
) -> Iterator[Tuple[str, np.ndarray]]:
    for task, (arr, lt) in labels.items():
        if lt == label_type:
            yield task, arr


def visualize(
    image: np.ndarray, labels: Labels, class_names: Dict[str, List[str]]
) -> np.ndarray:
    """Visualizes the labels on the image.

    @type image: np.ndarray
    @param image: The image to visualize.
    @type labels: Labels
    @param labels: The labels to visualize.
    @type class_names: Dict[str, List[str]]
    @param class_names: A dictionary mapping task names to a list of
        class names.
    @rtype: np.ndarray
    @return: The visualized image.
    """
    h, w, _ = image.shape
    images = {"image": image}

    for task, arr in _label_type_iterator(labels, LabelType.BOUNDINGBOX):
        curr_image = image.copy()
        for box in arr:
            cv2.rectangle(
                curr_image,
                (int(box[1] * w), int(box[2] * h)),
                (int(box[1] * w + box[3] * w), int(box[2] * h + box[4] * h)),
                _task_to_rgb(class_names[task][int(box[0])]),
                2,
            )
        images[task] = curr_image

    for task, arr in _label_type_iterator(labels, LabelType.KEYPOINTS):
        curr_image = image.copy()
        task_classes = class_names[task]

        *prefix, suffix = task.split("-")
        prefix = "".join(prefix)
        if suffix == "keypoints":
            if not prefix:
                bbox_task = "boundingbox"
            else:
                bbox_task = f"{prefix}-boundingbox"
            if bbox_task in images:
                curr_image = images.pop(bbox_task)
                if prefix:
                    task = f"{prefix} (keypoints + boundingbox)"
                else:
                    task = "keypoints + boundingbox"
        for kp in arr:
            cls_ = int(kp[0])
            kp = kp[1:].reshape(-1, 3)
            for k in kp:
                cv2.circle(
                    curr_image,
                    (int(k[0] * w), int(k[1] * h)),
                    2,
                    _task_to_rgb(task_classes[cls_]),
                    2,
                )
        images[task] = curr_image

    for task, arr in _label_type_iterator(labels, LabelType.SEGMENTATION):
        mask_viz = np.zeros((h, w, 3)).astype(np.uint8)
        for i, mask in enumerate(arr):
            mask = cv2.resize(mask, (w, h), interpolation=cv2.INTER_NEAREST)
            mask_viz[mask == 1] = (
                _task_to_rgb(class_names[task][i])
                if (i != 0 or len(arr) == 1)
                else (0, 0, 0)
            )
        images[task] = mask_viz

    return concat_images(images)



import random
from typing import Any, Dict, List, Tuple, Union

import albumentations as A
import cv2
import numpy as np
from albumentations import BoxType, KeypointType

from ..batch_transform import BatchBasedTransform
from ..utils import AUGMENTATIONS


@AUGMENTATIONS.register_module()
class MixUp(BatchBasedTransform):
    def __init__(
        self,
        alpha: Union[float, Tuple[float, float]] = 0.5,
        out_batch_size: int = 1,
        always_apply: bool = False,
        p: float = 0.5,
    ):
        """MixUp augmentation that merges two images and their
        annotations into one. If images are not of same size then second
        one is first resized to match the first one.

        @type alpha: Union[float, Tuple[float, float]]
        @param alpha: Mixing coefficient, either a single float or a
            tuple representing the range. Defaults to C{0.5}.
        @type out_batch_size: int
        @param out_batch_size: Number of output images in the batch.
            Defaults to C{1}.
        @type always_apply: bool
        @param always_apply: Whether to always apply the transform.
            Defaults to C{False}.
        @type p: float, optional
        @param p: Probability of applying the transform. Defaults to
            C{0.5}.
        """
        super().__init__(batch_size=2, always_apply=always_apply, p=p)

        self.alpha = alpha
        self.out_batch_size = out_batch_size

    def get_transform_init_args_names(self) -> Tuple[str, ...]:
        """Gets the default arguments for the mixup augmentation.

        @rtype: Tuple[str, ...]
        @return: The string keywords of the arguments.
        """
        return ("alpha", "out_batch_size")

    @property
    def targets_as_params(self) -> List[str]:
        """List of augmentation targets.

        @rtype: List[str]
        @return: Output list of augmentation targets.
        """
        return ["image_batch"]

    def apply_to_image_batch(
        self,
        image_batch: List[np.ndarray],
        image_shapes: List[Tuple[int, int]],
        **params,
    ) -> List[np.ndarray]:
        """Applies the transformation to a batch of images.

        @type image_batch: List[np.ndarray]
        @param image_batch: Batch of input images to which the
            transformation is applied.
        @type image_shapes: List[Tuple[int, int]]
        @param image_shapes: Shapes of the input images in the batch.
        @type params: Any
        @param params: Additional parameters for the transformation.
        @rtype: List[np.ndarray]
        @return: List of transformed images.
        """
        image1 = image_batch[0]
        # resize second image to size of the first one
        image2 = cv2.resize(
            image_batch[1], (image_shapes[0][1], image_shapes[0][0])
        )

        if isinstance(self.alpha, float):
            curr_alpha = np.clip(self.alpha, 0, 1)
        else:
            curr_alpha = random.uniform(
                max(self.alpha[0], 0), min(self.alpha[1], 1)
            )
        img_out = cv2.addWeighted(
            image1, curr_alpha, image2, 1 - curr_alpha, 0.0
        )
        return [img_out]

    def apply_to_mask_batch(
        self,
        mask_batch: List[np.ndarray],
        image_shapes: List[Tuple[int, int]],
        **params,
    ) -> List[np.ndarray]:
        """Applies the transformation to a batch of masks.

        @type mask_batch: List[np.ndarray]
        @param mask_batch: Batch of input masks to which the
            transformation is applied.
        @type image_shapes: List[Tuple[int, int]]
        @param image_shapes: Shapes of the input images in the batch.
        @type params: Any
        @param params: Additional parameters for the transformation.
        @rtype: List[np.ndarray]
        @return: List of transformed masks.
        """
        mask1 = mask_batch[0]
        mask2 = cv2.resize(
            mask_batch[1],
            (image_shapes[0][1], image_shapes[0][0]),
            interpolation=cv2.INTER_NEAREST,
        )
        out_mask = mask1 + mask2
        # if masks intersect keep one present in first image
        mask_inter = mask1 > 0
        out_mask[mask_inter] = mask1[mask_inter]
        return [out_mask]

    def apply_to_bboxes_batch(
        self,
        bboxes_batch: List[BoxType],
        image_shapes: List[Tuple[int, int]],
        **kwargs,
    ) -> List[BoxType]:
        """Applies the transformation to a batch of bboxes.

        @type bboxes_batch: List[BoxType]
        @param bboxes_batch: Batch of input bboxes to which the
            transformation is applied.
        @type image_shapes: List[Tuple[int, int]]
        @param image_shapes: Shapes of the input images in the batch.
        @type kwargs: Any
        @param kwargs: Additional parameters for the transformation.
        @rtype: List[BoxType]
        @return: List of transformed bboxes.
        """
        return [bboxes_batch[0] + bboxes_batch[1]]

    def apply_to_keypoints_batch(
        self,
        keypoints_batch: List[KeypointType],
        image_shapes: List[Tuple[int, int]],
        **kwargs,
    ) -> List[KeypointType]:
        """Applies the transformation to a batch of keypoints.

        @type keypoints_batch: List[BoxType]
        @param keypoints_batch: Batch of input keypoints to which the
            transformation is applied.
        @type image_shapes: List[Tuple[int, int]]
        @param image_shapes: Shapes of the input images in the batch.
        @type kwargs: Any
        @param kwargs: Additional parameters for the transformation.
        @rtype: List[BoxType]
        @return: List of transformed keypoints.
        """
        scaled_kpts2 = []
        scale_x = image_shapes[0][1] / image_shapes[1][1]
        scale_y = image_shapes[0][0] / image_shapes[1][0]
        for kpt in keypoints_batch[1]:
            new_kpt = A.augmentations.geometric.functional.keypoint_scale(
                keypoint=kpt, scale_x=scale_x, scale_y=scale_y
            )
            scaled_kpts2.append(new_kpt + kpt[4:])
        return [keypoints_batch[0] + scaled_kpts2]

    def get_params_dependent_on_targets(
        self, params: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Get parameters dependent on the targets.

        @param params: Dictionary containing parameters.
        @type params: Dict[str, Any]
        @return: Dictionary containing parameters dependent on the
            targets.
        @rtype: Dict[str, Any]
        """
        image_batch = params["image_batch"]
        return {"image_shapes": [image.shape[:2] for image in image_batch]}



import copy
from typing import Any, Dict, Optional, Sequence

from albumentations.core.bbox_utils import BboxParams, BboxProcessor
from albumentations.core.keypoints_utils import (
    KeypointParams,
    KeypointsProcessor,
)
from albumentations.core.utils import DataProcessor

from .batch_utils import batch2list, list2batch, to_unbatched_name


class BboxBatchProcessor(DataProcessor):
    def __init__(
        self,
        params: BboxParams,
        additional_targets: Optional[Dict[str, str]] = None,
    ):
        """Data processor class to process bbox data in batches.

        @param params: Bbox parameters
        @type params: BboxParams
        @param additional_targets: Additional targets of the transform.
            Defaults to None.
        @type additional_targets: Optional[Dict[str, str]]
        """
        super().__init__(params, additional_targets)
        item_params = copy.deepcopy(params)
        if item_params.label_fields is not None:
            label_fields = item_params.label_fields
            item_params.label_fields = [
                to_unbatched_name(field) for field in label_fields
            ]
        self.item_processor = BboxProcessor(item_params, additional_targets)

    @property
    def default_data_name(self) -> str:
        return "bboxes_batch"

    def ensure_data_valid(self, data: Dict[str, Any]) -> None:
        for item in batch2list(data):
            self.item_processor.ensure_data_valid(item)

    def postprocess(self, data: Dict[str, Any]) -> Dict[str, Any]:
        processed = [
            self.item_processor.postprocess(item) for item in batch2list(data)
        ]
        procesed_data = list2batch(processed)
        for k in data.keys():
            data[k] = procesed_data[k]
        return data

    def preprocess(self, data: Dict[str, Any]) -> None:
        processed = batch2list(data)
        for item in processed:
            self.item_processor.preprocess(item)
        procesed_data = list2batch(processed)
        for k in data.keys():
            data[k] = procesed_data[k]

    def filter_batch(self, batched_data: Dict[str, Any]) -> Dict[str, Any]:
        processed = []
        for data in batch2list(batched_data):
            rows, cols = data["image"][:2]
            for data_name in self.item_processor.data_fields:
                data[data_name] = self.item_processor.filter(
                    data[data_name], rows, cols
                )
            processed.append(data)
        return list2batch(processed)

    def filter(self, data: Sequence, rows: int, cols: int) -> Sequence:
        return self.item_processor.filter(data, rows, cols)

    def check(self, data: Sequence, rows: int, cols: int) -> None:
        return self.item_processor.check(data, rows, cols)

    def convert_to_albumentations(
        self, data: Sequence, rows: int, cols: int
    ) -> Sequence:
        return self.item_processor.convert_to_albumentations(data, rows, cols)

    def convert_from_albumentations(
        self, data: Sequence, rows: int, cols: int
    ) -> Sequence:
        return self.item_processor.convert_from_albumentations(
            data, rows, cols
        )


class KeypointsBatchProcessor(DataProcessor):
    def __init__(
        self,
        params: KeypointParams,
        additional_targets: Optional[Dict[str, str]] = None,
    ):
        """Data processor class to process keypoint data in batches.

        @param params: Keypoint parameters
        @type params: KeypointParams
        @param additional_targets: Additional targets of the transform.
            Defaults to None.
        @type additional_targets: Optional[Dict[str, str]]
        """
        super().__init__(params, additional_targets)
        item_params = copy.deepcopy(params)
        if item_params.label_fields is not None:
            label_fields = item_params.label_fields
            item_params.label_fields = [
                to_unbatched_name(field) for field in label_fields
            ]
        self.item_processor = KeypointsProcessor(
            item_params, additional_targets
        )

    @property
    def default_data_name(self) -> str:
        return "keypoints_batch"

    def ensure_data_valid(self, data: Dict[str, Any]) -> None:
        for item in batch2list(data):
            self.item_processor.ensure_data_valid(item)

    def postprocess(self, data: Dict[str, Any]) -> Dict[str, Any]:
        processed = [
            self.item_processor.postprocess(item) for item in batch2list(data)
        ]
        procesed_data = list2batch(processed)
        for k in data.keys():
            data[k] = procesed_data[k]
        return data

    def preprocess(self, data: Dict[str, Any]) -> None:
        processed = batch2list(data)
        for item in processed:
            self.item_processor.preprocess(item)
        procesed_data = list2batch(processed)
        for k in data.keys():
            data[k] = procesed_data[k]

    def filter_batch(self, batched_data: Dict[str, Any]) -> Dict[str, Any]:
        processed = []
        for data in batch2list(batched_data):
            rows, cols = data["image"][:2]
            for data_name in self.item_processor.data_fields:
                data[data_name] = self.item_processor.filter(
                    data[data_name], rows, cols
                )
            processed.append(data)
        return list2batch(processed)

    def filter(self, data: Sequence, rows: int, cols: int) -> Sequence:
        return self.item_processor.filter(data, rows, cols)

    def check(self, data: Sequence, rows: int, cols: int) -> None:
        return self.item_processor.check(data, rows, cols)

    def convert_to_albumentations(
        self, data: Sequence, rows: int, cols: int
    ) -> Sequence:
        return self.item_processor.convert_to_albumentations(data, rows, cols)

    def convert_from_albumentations(
        self, data: Sequence, rows: int, cols: int
    ) -> Sequence:
        return self.item_processor.convert_from_albumentations(
            data, rows, cols
        )



import json
import logging
import math
import shutil
import tempfile
from collections import defaultdict
from contextlib import suppress
from pathlib import Path
from typing import (
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    overload,
)

import numpy as np
import polars as pl
import pyarrow.parquet as pq
from ordered_set import OrderedSet
from typing_extensions import Self

import luxonis_ml.data.utils.data_utils as data_utils
from luxonis_ml.utils import (
    LuxonisFileSystem,
    deprecated,
    environ,
    make_progress_bar,
)
from luxonis_ml.utils.filesystem import PathType

from ..utils.constants import LDF_VERSION
from ..utils.enums import BucketStorage, BucketType
from ..utils.parquet import ParquetFileManager
from .annotation import Annotation, ArrayAnnotation, DatasetRecord
from .base_dataset import BaseDataset, DatasetIterator
from .source import LuxonisSource
from .utils import add_generator_wrapper, find_filepath_uuid, get_dir, get_file


class LuxonisDataset(BaseDataset):
    def __init__(
        self,
        dataset_name: str,
        team_id: Optional[str] = None,
        bucket_type: BucketType = BucketType.INTERNAL,
        bucket_storage: BucketStorage = BucketStorage.LOCAL,
        *,
        delete_existing: bool = False,
        delete_remote: bool = False,
    ) -> None:
        """Luxonis Dataset Format (LDF) is used to define datasets in
        the Luxonis MLOps ecosystem.

        @type dataset_name: str
        @param dataset_name: Name of the dataset
        @type team_id: Optional[str]
        @param team_id: Optional unique team identifier for the cloud
        @type bucket_type: BucketType
        @param bucket_type: Whether to use external cloud buckets
        @type bucket_storage: BucketStorage
        @param bucket_storage: Underlying bucket storage. Can be one of
            C{local}, C{S3}, or C{GCS}.
        @type delete_existing: bool
        @param delete_existing: Whether to delete a dataset with the
            same name if it exists
        @type delete_remote: bool
        @param delete_remote: Whether to delete the dataset from the
            cloud as well
        """

        self.base_path = environ.LUXONISML_BASE_PATH
        self.base_path.mkdir(exist_ok=True)

        self._credentials = self._init_credentials()
        self._is_synced = False

        self.bucket_type = bucket_type
        self.bucket_storage = bucket_storage

        if self.bucket_storage == BucketStorage.AZURE_BLOB:
            raise NotImplementedError("Azure Blob Storage not yet supported")

        self.bucket = self._get_credential("LUXONISML_BUCKET")

        if self.is_remote and self.bucket is None:
            raise ValueError("Must set LUXONISML_BUCKET environment variable!")

        self.dataset_name = dataset_name
        self.team_id = team_id or self._get_credential("LUXONISML_TEAM_ID")

        if delete_existing:
            if LuxonisDataset.exists(
                dataset_name, team_id, bucket_storage, self.bucket
            ):
                LuxonisDataset(
                    dataset_name, team_id, bucket_type, bucket_storage
                ).delete_dataset(delete_remote=delete_remote)

        self._init_paths()

        if not self.is_remote:
            self.fs = LuxonisFileSystem(f"file://{self.path}")
        else:
            self.fs = LuxonisFileSystem(self.path)

        self.metadata = defaultdict(dict, self._get_metadata())

        self.logger = logging.getLogger(__name__)

        self.progress = make_progress_bar()

    @property
    def source(self) -> LuxonisSource:
        if "source" not in self.metadata:
            raise ValueError("Source not found in metadata")
        return LuxonisSource.from_document(self.metadata["source"])

    @property
    def identifier(self) -> str:
        return self.dataset_name

    def __len__(self) -> int:
        """Returns the number of instances in the dataset."""

        df = self._load_df_offline()
        return len(df.select("uuid").unique()) if df is not None else 0

    def _get_credential(self, key: str) -> str:
        """Gets secret credentials from credentials file or ENV
        variables."""

        if key in self._credentials.keys():
            return self._credentials[key]
        else:
            if not hasattr(environ, key):
                raise RuntimeError(f"Must set {key} in ENV variables")
            return getattr(environ, key)

    def _init_paths(self) -> None:
        """Configures local path or bucket directory."""

        self.local_path = (
            self.base_path
            / "data"
            / self.team_id
            / "datasets"
            / self.dataset_name
        )
        self.media_path = self.local_path / "media"
        self.annotations_path = self.local_path / "annotations"
        self.metadata_path = self.local_path / "metadata"
        self.arrays_path = self.local_path / "arrays"

        for path in [
            self.media_path,
            self.annotations_path,
            self.metadata_path,
        ]:
            path.mkdir(exist_ok=True, parents=True)

        if not self.is_remote:
            self.path = str(self.local_path)
        else:
            self.path = self._construct_url(
                self.bucket_storage,
                self.bucket,
                self.team_id,
                self.dataset_name,
            )

    @overload
    def _load_df_offline(
        self, lazy: Literal[False] = ...
    ) -> Optional[pl.DataFrame]: ...

    @overload
    def _load_df_offline(
        self, lazy: Literal[True] = ...
    ) -> Optional[pl.LazyFrame]: ...

    def _load_df_offline(
        self, lazy: bool = False
    ) -> Optional[Union[pl.DataFrame, pl.LazyFrame]]:
        path = get_dir(self.fs, "annotations", self.local_path)

        if path is None or not path.exists():
            return None

        if lazy:
            dfs = [pl.scan_parquet(file) for file in path.glob("*.parquet")]
            return pl.concat(dfs) if dfs else None
        else:
            dfs = [pl.read_parquet(file) for file in path.glob("*.parquet")]
            return pl.concat(dfs) if dfs else None

    @overload
    def _get_file_index(
        self, lazy: Literal[False] = ...
    ) -> Optional[pl.DataFrame]: ...

    @overload
    def _get_file_index(
        self, lazy: Literal[True] = ...
    ) -> Optional[pl.LazyFrame]: ...

    def _get_file_index(
        self, lazy: bool = False
    ) -> Optional[Union[pl.DataFrame, pl.LazyFrame]]:
        path = get_file(
            self.fs, "metadata/file_index.parquet", self.media_path
        )
        if path is not None and path.exists():
            if not lazy:
                df = pl.read_parquet(path)
            else:
                df = pl.scan_parquet(path)

            return df.select(pl.all().exclude("^__index_level_.*$"))

        return None

    def _write_index(
        self,
        index: Optional[pl.DataFrame],
        new_index: Dict[str, List[str]],
        path: Optional[PathType] = None,
    ) -> None:
        path = Path(path or self.metadata_path / "file_index.parquet")
        df = pl.DataFrame(new_index)
        if index is not None:
            df = pl.concat([index, df])
        pq.write_table(df.to_arrow(), path)

    def _write_metadata(self) -> None:
        path = self.metadata_path / "metadata.json"
        path.write_text(json.dumps(self.metadata, indent=4))
        with suppress(shutil.SameFileError):
            self.fs.put_file(path, "metadata/metadata.json")

    @staticmethod
    def _construct_url(
        bucket_storage: BucketStorage,
        bucket: str,
        team_id: str,
        dataset_name: str,
    ) -> str:
        """Constructs a URL for a remote dataset."""
        return f"{bucket_storage.value}://{bucket}/{team_id}/datasets/{dataset_name}"

    def _init_credentials(self) -> Dict[str, Any]:
        credentials_cache_file = self.base_path / "credentials.json"
        if credentials_cache_file.exists():
            return json.loads(credentials_cache_file.read_text())
        return {}

    def _get_metadata(self) -> Dict[str, Any]:
        if self.fs.exists("metadata/metadata.json"):
            path = get_file(
                self.fs,
                "metadata/metadata.json",
                self.metadata_path,
                default=self.metadata_path / "metadata.json",
            )
            return json.loads(path.read_text())
        else:
            return {
                "source": LuxonisSource().to_document(),
                "ldf_version": LDF_VERSION,
                "classes": {},
            }

    @property
    def is_remote(self) -> bool:
        return self.bucket_storage != BucketStorage.LOCAL

    def update_source(self, source: LuxonisSource) -> None:
        """Updates underlying source of the dataset with a new
        L{LuxonisSource}.

        @type source: LuxonisSource
        @param source: The new C{LuxonisSource} to replace the old one.
        """

        self.metadata["source"] = source.to_document()
        self._write_metadata()

    def set_classes(
        self, classes: List[str], task: Optional[str] = None
    ) -> None:
        if task is not None:
            self.metadata["classes"][task] = classes
        else:
            raise NotImplementedError(
                "Setting classes for all tasks not yet supported. "
                "Set classes individually for each task"
            )
        self._write_metadata()

    def get_classes(self) -> Tuple[List[str], Dict[str, List[str]]]:
        all_classes = list(
            {
                c
                for classes in self.metadata["classes"].values()
                for c in classes
            }
        )
        for task in self.metadata["classes"]:
            self.metadata["classes"][task] = sorted(
                self.metadata["classes"][task]
            )
        return sorted(all_classes), self.metadata["classes"]

    def set_skeletons(
        self,
        labels: Optional[List[str]] = None,
        edges: Optional[List[Tuple[int, int]]] = None,
        task: Optional[str] = None,
    ) -> None:
        if labels is None and edges is None:
            raise ValueError("Must provide either keypoint names or edges")

        if task is None:
            tasks = self.get_tasks()
        else:
            tasks = [task]
        for task in tasks:
            self.metadata["skeletons"][task] = {
                "labels": labels or [],
                "edges": edges or [],
            }
        self._write_metadata()

    def get_skeletons(
        self,
    ) -> Dict[str, Tuple[List[str], List[Tuple[int, int]]]]:
        return {
            task: (skel["labels"], skel["edges"])
            for task, skel in self.metadata["skeletons"].items()
        }

    def get_tasks(self) -> List[str]:
        return list(self.get_classes()[1].keys())

    def sync_from_cloud(self, force: bool = False) -> None:
        """Downloads data from a remote cloud bucket."""

        if not self.is_remote:
            self.logger.warning("This is a local dataset! Cannot sync")
        else:
            if not self._is_synced or force:
                self.logger.info("Syncing from cloud...")
                local_dir = self.base_path / "data" / self.team_id / "datasets"
                local_dir.mkdir(exist_ok=True, parents=True)

                self.fs.get_dir(remote_paths="", local_dir=local_dir)

                self._is_synced = True
            else:
                self.logger.warning("Already synced. Use force=True to resync")

    def delete_dataset(self, *, delete_remote: bool = False) -> None:
        """Deletes the dataset from local storage and optionally from
        the cloud.

        @type delete_remote: bool
        @param delete_remote: Whether to delete the dataset from the
            cloud.
        """
        if not self.is_remote:
            shutil.rmtree(self.path)
            self.logger.info(f"Deleted dataset {self.dataset_name}")

        if self.is_remote and delete_remote:
            self.logger.info(
                f"Deleting dataset {self.dataset_name} from cloud"
            )
            assert self.path
            assert self.dataset_name
            assert self.local_path
            if self.local_path.exists():
                shutil.rmtree(self.local_path)
            self.fs.delete_dir(allow_delete_parent=True)

    def _infer_task(self, ann: Annotation) -> str:
        if not hasattr(LuxonisDataset._infer_task, "_logged_infered_classes"):
            LuxonisDataset._infer_task._logged_infered_classes = defaultdict(
                bool
            )

        def _log_once(cls_: str, task: str, message: str, level: str = "info"):
            if not LuxonisDataset._infer_task._logged_infered_classes[
                (cls_, task)
            ]:
                LuxonisDataset._infer_task._logged_infered_classes[
                    (cls_, task)
                ] = True
                getattr(self.logger, level)(message, extra={"markup": True})

        cls_ = ann.class_
        _, current_classes = self.get_classes()
        infered_task = None

        for task, classes in current_classes.items():
            if cls_ in classes:
                if infered_task is not None:
                    _log_once(
                        cls_,
                        infered_task,
                        f"Class [red italic]{cls_}[reset] is ambiguous between "
                        "tasks [magenta italic]{infered_task}[reset] and [magenta italic]{task}[reset]. "
                        "Task inference failed.",
                        "warning",
                    )
                    infered_task = None
                    break
                infered_task = task
        if infered_task is None:
            _log_once(
                cls_,
                ann.task,
                f"Class [red italic]{cls_}[reset] doesn't belong to any existing task. "
                f"Autogenerated task [magenta italic]{ann.task}[reset] will be used.",
                "info",
            )
        else:
            _log_once(
                cls_,
                infered_task,
                f"Class [red italic]{cls_}[reset] infered to belong to task [magenta italic]{infered_task}[reset]",
            )
            return infered_task

        return ann.task

    def _process_arrays(self, batch_data: List[DatasetRecord]) -> None:
        self.logger.info("Checking arrays...")
        task = self.progress.add_task(
            "[magenta]Processing arrays...", total=len(batch_data)
        )
        self.progress.start()
        uuid_dict = {}
        for record in batch_data:
            self.progress.update(task, advance=1)
            if not isinstance(record.annotation, ArrayAnnotation):
                continue
            ann = record.annotation
            data_utils.check_array(ann.path)
            if self.is_remote:
                uuid = self.fs.get_file_uuid(
                    ann.path, local=True
                )  # TODO: support from bucket
                uuid_dict[str(ann.path)] = uuid
                ann.path = Path(uuid).with_suffix(ann.path.suffix)
            else:
                ann.path = ann.path.absolute().resolve()
        self.progress.stop()
        self.progress.remove_task(task)
        if self.is_remote:
            self.logger.info("Uploading arrays...")
            # TODO: support from bucket (likely with a self.fs.copy_dir)
            self.fs.put_dir(
                local_paths=uuid_dict.keys(),
                remote_dir="arrays",
                uuid_dict=uuid_dict,
            )

    def _add_process_batch(
        self,
        batch_data: List[DatasetRecord],
        pfm: ParquetFileManager,
        index: Optional[pl.DataFrame],
        new_index: Dict[str, List[str]],
        processed_uuids: Set[str],
    ) -> None:
        paths = list(set(data.file for data in batch_data))
        self.logger.info("Generating UUIDs...")
        uuid_dict = self.fs.get_file_uuids(
            paths, local=True
        )  # TODO: support from bucket
        if self.is_remote:
            self.logger.info("Uploading media...")

            # TODO: support from bucket (likely with a self.fs.copy_dir)
            self.fs.put_dir(
                local_paths=paths, remote_dir="media", uuid_dict=uuid_dict
            )
            self.logger.info("Media uploaded")

        self._process_arrays(batch_data)

        task = self.progress.add_task(
            "[magenta]Processing data...", total=len(batch_data)
        )

        self.logger.info("Saving annotations...")
        with self.progress:
            for ann in batch_data:
                filepath = ann.file
                file = filepath.name
                uuid = uuid_dict[str(filepath)]
                matched_id = find_filepath_uuid(filepath, index)
                if matched_id is not None:
                    if matched_id != uuid:
                        # TODO: not sure if this should be an exception or how we should really handle it
                        raise ValueError(
                            f"{filepath} already added to the dataset! Please skip or rename the file."
                        )
                        # TODO: we may also want to check for duplicate uuids to get a one-to-one relationship
                elif uuid not in processed_uuids:
                    new_index["uuid"].append(uuid)
                    new_index["file"].append(file)
                    new_index["original_filepath"].append(
                        str(filepath.absolute().resolve())
                    )
                    processed_uuids.add(uuid)

                pfm.write({"uuid": uuid, **ann.to_parquet_dict()})
                self.progress.update(task, advance=1)
        self.progress.remove_task(task)

    def add(
        self, generator: DatasetIterator, batch_size: int = 1_000_000
    ) -> Self:
        generator = add_generator_wrapper(generator)
        index = self._get_file_index()
        new_index = {"uuid": [], "file": [], "original_filepath": []}
        processed_uuids = set()

        batch_data: list[DatasetRecord] = []

        classes_per_task: Dict[str, OrderedSet[str]] = defaultdict(
            lambda: OrderedSet([])
        )
        num_kpts_per_task: Dict[str, int] = {}

        annotations_path = get_dir(
            self.fs,
            "annotations",
            self.local_path,
            default=self.annotations_path,
        )
        assert annotations_path is not None

        with ParquetFileManager(annotations_path) as pfm:
            for i, data in enumerate(generator, start=1):
                record = (
                    data
                    if isinstance(data, DatasetRecord)
                    else DatasetRecord(**data)
                )
                ann = record.annotation
                if ann is not None:
                    if ann.task == ann._label_type.value and (
                        isinstance(data, DatasetRecord)
                        or "task" not in data.get("annotation", {})
                    ):
                        ann.task = self._infer_task(ann)

                    classes_per_task[ann.task].add(ann.class_)
                    if ann.type_ == "keypoints":
                        num_kpts_per_task[ann.task] = len(ann.keypoints)

                batch_data.append(record)
                if i % batch_size == 0:
                    self._add_process_batch(
                        batch_data, pfm, index, new_index, processed_uuids
                    )
                    batch_data = []

            self._add_process_batch(
                batch_data, pfm, index, new_index, processed_uuids
            )

        with suppress(shutil.SameFileError):
            self.fs.put_dir(annotations_path, "")

        _, curr_classes = self.get_classes()
        for task, classes in classes_per_task.items():
            old_classes = set(curr_classes.get(task, []))
            new_classes = list(classes - old_classes)
            if new_classes:
                self.logger.info(
                    f"Detected new classes for task {task}: {new_classes}"
                )
                self.set_classes(list(classes | old_classes), task)
        for task, num_kpts in num_kpts_per_task.items():
            self.set_skeletons(
                labels=[str(i) for i in range(num_kpts)],
                edges=[(i, i + 1) for i in range(num_kpts - 1)],
                task=task,
            )

        with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
            self._write_index(index, new_index, path=tmp_file.name)

        self.fs.put_file(tmp_file.name, "metadata/file_index.parquet")
        self._write_metadata()
        self._warn_on_duplicates()
        return self

    def _warn_on_duplicates(self) -> None:
        df = self._load_df_offline(lazy=True)
        index_df = self._get_file_index(lazy=True)
        if df is None or index_df is None:
            return
        df = df.join(index_df, on="uuid").drop("file_right")
        # Warn on duplicate UUIDs
        duplicates_paired = (
            df.group_by("uuid")
            .agg(pl.col("file").n_unique().alias("file_count"))
            .filter(pl.col("file_count") > 1)
            .join(df, on="uuid")
            .select("uuid", "file")
            .unique()
            .group_by("uuid")
            .agg(pl.col("file").alias("files"))
            .filter(pl.col("files").len() > 1)
            .collect()
        )
        for uuid, files in duplicates_paired.iter_rows():
            self.logger.warning(
                f"UUID: {uuid} has multiple file names: {files}"
            )

        # Warn on duplicate annotations
        duplicate_annotation = (
            df.group_by(
                "original_filepath",
                "task",
                "type",
                "annotation",
                "instance_id",
            )
            .agg(pl.len().alias("count"))
            .filter(pl.col("count") > 1)
            .filter(pl.col("annotation") != "{}")
            .drop("instance_id")
        ).collect()

        for (
            file_name,
            task,
            type_,
            annotation,
            count,
        ) in duplicate_annotation.iter_rows():
            if "RLE" in type_ or "Mask" in type_:
                annotation = "<binary mask>"
            self.logger.warning(
                f"File '{file_name}' has the same '{type_}' annotation "
                f"'{annotation}' ({task=}) added {count} times."
            )

    def get_splits(self) -> Optional[Dict[str, List[str]]]:
        splits_path = get_file(
            self.fs,
            "metadata/splits.json",
            self.metadata_path,
        )
        if splits_path is None:
            return None

        with open(splits_path, "r") as file:
            return json.load(file)

    @deprecated(
        "ratios",
        "definitions",
        suggest={"ratios": "splits", "definitions": "splits"},
    )
    def make_splits(
        self,
        splits: Optional[
            Union[
                Dict[str, Sequence[PathType]],
                Dict[str, float],
                Tuple[float, float, float],
            ]
        ] = None,
        *,
        ratios: Optional[
            Union[Dict[str, float], Tuple[float, float, float]]
        ] = None,
        definitions: Optional[Dict[str, List[PathType]]] = None,
        replace_old_splits: bool = False,
    ) -> None:
        if ratios is not None and definitions is not None:
            raise ValueError("Cannot provide both ratios and definitions")

        if splits is None and ratios is None and definitions is None:
            splits = {"train": 0.8, "val": 0.1, "test": 0.1}

        if splits is not None:
            if ratios is not None or definitions is not None:
                raise ValueError(
                    "Cannot provide both splits and ratios/definitions"
                )
            if isinstance(splits, tuple):
                ratios = splits
            elif isinstance(splits, dict):
                value = next(iter(splits.values()))
                if isinstance(value, float):
                    ratios = splits  # type: ignore
                elif isinstance(value, list):
                    definitions = splits  # type: ignore

        if ratios is not None:
            if isinstance(ratios, tuple):
                if not len(ratios) == 3:
                    raise ValueError(
                        "Ratios must be a tuple of 3 floats for train, val, and test splits"
                    )
                ratios = {
                    "train": ratios[0],
                    "val": ratios[1],
                    "test": ratios[2],
                }
            sum_ = sum(ratios.values())
            if not math.isclose(sum_, 1.0):
                raise ValueError(f"Ratios must sum to 1.0, got {sum_:0.4f}")

        if definitions is not None:
            n_files = sum(map(len, definitions.values()))
            if n_files > len(self):
                raise ValueError(
                    "Dataset size is smaller than the total number of files in the definitions. "
                    f"Dataset size: {len(self)}, Definitions: {n_files}."
                )

        splits_to_update: List[str] = []
        new_splits: Dict[str, List[str]] = {}
        old_splits: Dict[str, List[str]] = defaultdict(list)

        splits_path = get_file(
            self.fs,
            "metadata/splits.json",
            self.metadata_path,
            default=self.metadata_path / "splits.json",
        )
        if splits_path.exists():
            with open(splits_path, "r") as file:
                old_splits = defaultdict(list, json.load(file))

        defined_uuids = set(
            uuid for uuids in old_splits.values() for uuid in uuids
        )

        if definitions is None:
            ratios = ratios or {"train": 0.8, "val": 0.1, "test": 0.1}
            df = self._load_df_offline()
            assert df is not None
            ids = (
                df.filter(~pl.col("uuid").is_in(defined_uuids))
                .select("uuid")
                .unique()
                .get_column("uuid")
                .to_list()
            )
            if not ids:
                if not replace_old_splits:
                    raise ValueError(
                        "No new files to add to splits. "
                        "If you want to generate new splits, set `replace_old_splits=True`"
                    )
                else:
                    ids = (
                        df.select("uuid").unique().get_column("uuid").to_list()
                    )
                    old_splits = defaultdict(list)

            np.random.shuffle(ids)
            N = len(ids)
            lower_bound = 0
            for split, ratio in ratios.items():
                upper_bound = lower_bound + math.ceil(N * ratio)
                new_splits[split] = ids[lower_bound:upper_bound]
                splits_to_update.append(split)
                lower_bound = upper_bound

        else:
            index = self._get_file_index()
            if index is None:
                raise FileNotFoundError("File index not found")
            for split, filepaths in definitions.items():
                splits_to_update.append(split)
                if not isinstance(filepaths, list):
                    raise ValueError(
                        "Must provide splits as a list of filepaths"
                    )
                ids = [
                    find_filepath_uuid(filepath, index, raise_on_missing=True)
                    for filepath in filepaths
                ]
                new_splits[split] = ids

        for split, uuids in new_splits.items():
            old_splits[split].extend(uuids)

        splits_path.write_text(json.dumps(old_splits, indent=4))

        with suppress(shutil.SameFileError):
            self.fs.put_file(splits_path, "metadata/splits.json")

    @staticmethod
    def exists(
        dataset_name: str,
        team_id: Optional[str] = None,
        bucket_storage: BucketStorage = BucketStorage.LOCAL,
        bucket: Optional[str] = None,
    ) -> bool:
        """Checks if a dataset exists.

        @type dataset_name: str
        @param dataset_name: Name of the dataset to check
        @type team_id: Optional[str]
        @param team_id: Optional team identifier
        @type bucket_storage: BucketStorage
        @param bucket_storage: Underlying bucket storage from C{local},
            C{S3}, or C{GCS}. Default is C{local}.
        @type bucket: Optional[str]
        @param bucket: Name of the bucket. Default is C{None}.
        @rtype: bool
        @return: Whether the dataset exists.
        """
        return dataset_name in LuxonisDataset.list_datasets(
            team_id, bucket_storage, bucket
        )

    @staticmethod
    def list_datasets(
        team_id: Optional[str] = None,
        bucket_storage: BucketStorage = BucketStorage.LOCAL,
        bucket: Optional[str] = None,
    ) -> List[str]:
        """Returns a dictionary of all datasets.

        @type team_id: Optional[str]
        @param team_id: Optional team identifier
        @type bucket_storage: BucketStorage
        @param bucket_storage: Underlying bucket storage from C{local},
            C{S3}, or C{GCS}. Default is C{local}.
        @type bucket: Optional[str]
        @param bucket: Name of the bucket. Default is C{None}.
        @rtype: List[str]
        @return: List of all dataset names.
        """
        base_path = environ.LUXONISML_BASE_PATH

        team_id = team_id or environ.LUXONISML_TEAM_ID

        if bucket_storage == BucketStorage.LOCAL:
            local_path = base_path / "data" / team_id / "datasets"
            if not local_path.exists():
                return []
            return [d.name for d in local_path.iterdir() if d.is_dir()]

        bucket = bucket or environ.LUXONISML_BUCKET
        if bucket is None:
            raise ValueError("Must set LUXONISML_BUCKET environment variable!")

        fs = LuxonisFileSystem(
            LuxonisDataset._construct_url(bucket_storage, bucket, team_id, "")
        )
        return list(fs.walk_dir("", recursive=False, typ="directory"))
